#! /usr/bin/env python3
"""
Translate NSO Device config to MDD OpenConfig

This script will pull a device's configuration from an NSO server, convert the NED structured configuration to
MDD OpenConfig, save the NSO configuration to a file named {device_name}_configuration.json, save the NSO device
configuration minus parts replaced by OpenConfig to a file named {device_name}_configuration_remaining.json,
and save the MDD OpenConfig configuration to a file named {nso_device}_openconfig.json.

The script requires the following environment variables:
NSO_URL - URL for the NSO server
NSO_USERNAME
NSO_PASSWORD
NSO_DEVICE - NSO device name for configuration translation
TEST - True or False. True enables sending the OpenConfig to the NSO server after generation
"""

import sys
from pathlib import Path
from importlib.util import find_spec

system_notes = []

openconfig_system = {
    "openconfig-system:system": {
        "openconfig-system:aaa": {},
        "openconfig-system:clock": {},
        "openconfig-system:config": {},
        "openconfig-system:dns": {},
        "openconfig-system:logging": {},
        "openconfig-system:ntp": {
            "openconfig-system:config": {},
            "openconfig-system:ntp-keys": {
                "openconfig-system:ntp-key": []},
            "openconfig-system:servers": {
                "openconfig-system:server": []}
        },
        "openconfig-system:ssh-server": {
            "openconfig-system:config": {},
            "openconfig-system-ext:algorithm": {
                "openconfig-system-ext:config": {}
            }
        },
        "openconfig-system-ext:services": {"openconfig-system-ext:config": {}}
    }
}


def xr_system_services(config_before: dict, config_leftover: dict) -> None:
    """
    Translates NSO XR NED to MDD OpenConfig System Services
    """
    openconfig_system_services = openconfig_system["openconfig-system:system"]["openconfig-system-ext:services"]
    if type(config_before.get("tailf-ned-cisco-ios-xr:domain", {}).get("lookup", {}).get("disable", "")) is list:
        openconfig_system_services["openconfig-system-ext:config"]["openconfig-system-ext:ip-domain-lookup"] = False
        del config_leftover["tailf-ned-cisco-ios-xr:domain"]["lookup"]
    else:
        openconfig_system_services["openconfig-system-ext:config"]["openconfig-system-ext:ip-domain-lookup"] = True
    # service tcp-small-servers
    if type(config_before.get("tailf-ned-cisco-ios-xr:service", {}).get("ipv4", {}).get("tcp-small-servers", '')) is dict:
        openconfig_system_services["openconfig-system-ext:config"]["openconfig-system-ext:service-tcp-small-servers"] = True
        del config_leftover["tailf-ned-cisco-ios-xr:service"]["ipv4"]["tcp-small-servers"]["max-servers"]
    else:
        openconfig_system_services["openconfig-system-ext:config"]["openconfig-system-ext:service-tcp-small-servers"] = False
    # service udp-small-servers
    if type(config_before.get("tailf-ned-cisco-ios-xr:service", {}).get("ipv4", {}).get("udp-small-servers", '')) is dict:
        openconfig_system_services["openconfig-system-ext:config"]["openconfig-system-ext:service-udp-small-servers"] = True
        del config_leftover["tailf-ned-cisco-ios-xr:service"]["ipv4"]["udp-small-servers"]["max-servers"]
    else:
        openconfig_system_services["openconfig-system-ext:config"]["openconfig-system-ext:service-udp-small-servers"] = False

def xr_system_config(config_before: dict, config_leftover: dict) -> None:
    """
    Translates NSO XE NED to MDD OpenConfig System Config
    """
    openconfig_system_config = openconfig_system["openconfig-system:system"]["openconfig-system:config"]
    default_secret = config_before.get("tailf-ned-cisco-ios-xr:line", {}).get("default", {}).get("secret", {})
    console_exec_timeout = config_before.get("tailf-ned-cisco-ios-xr:line", {}).get("console", {}).get("exec-timeout", {})

    openconfig_system_config["openconfig-system:hostname"] = config_before["tailf-ned-cisco-ios-xr:hostname"]
    del config_leftover["tailf-ned-cisco-ios-xr:hostname"]

    if config_before.get("tailf-ned-cisco-ios-xr:banner", {}).get("login", {}).get("message"):
        openconfig_system_config["openconfig-system:login-banner"] = (
            config_before.get("tailf-ned-cisco-ios-xr:banner",{})
            .get("login")
            .get("message")
        )
        del config_leftover["tailf-ned-cisco-ios-xr:banner"]["login"]

    if config_before.get("tailf-ned-cisco-ios-xr:banner", {}).get("motd", {}).get("message"):
        openconfig_system_config["openconfig-system:motd-banner"] = (
            config_before.get("tailf-ned-cisco-ios-xr:banner",{})
            .get("motd")
            .get("message")
        )
        del config_leftover["tailf-ned-cisco-ios-xr:banner"]["motd"]

    if config_before.get("tailf-ned-cisco-ios-xr:domain", {}).get("name"):
        openconfig_system_config["openconfig-system:domain-name"] = (
            config_before.get("tailf-ned-cisco-ios-xr:domain", {})
            .get("name")
        )
        del config_leftover["tailf-ned-cisco-ios-xr:domain"]["name"]

    if default_secret.get("secret") and default_secret.get("type") == "0":
        openconfig_system_config["openconfig-system-ext:enable-secret"] = default_secret.get("secret")
        del config_leftover["tailf-ned-cisco-ios-xr:line"]["default"]["secret"]

    if console_exec_timeout.get('minutes') or console_exec_timeout.get('seconds'):
        seconds = console_exec_timeout.get('minutes') * 60
        seconds += console_exec_timeout.get('seconds', 0)
        openconfig_system_config["openconfig-system-ext:console-exec-timeout-seconds"] = seconds
        del config_leftover["tailf-ned-cisco-ios-xr:line"]["console"]["exec-timeout"]

def xr_system_ssh_server(config_before: dict, config_leftover: dict) -> None:
    """
    Translates NSO XR NED to MDD OpenConfig System SSH Server
    """
    openconfig_system_ssh_server_alg_config = openconfig_system["openconfig-system:system"]["openconfig-system:ssh-server"]["openconfig-system-ext:algorithm"]["openconfig-system-ext:config"]

    if type(config_before.get("tailf-ned-cisco-ios-xr:ssh", {}).get("server", {}).get("algorithms", {}).get("cipher", '')) is list:
        openconfig_system_ssh_server_alg_config["openconfig-system-ext:encryption"] = config_before.get("tailf-ned-cisco-ios-xr:ssh", {}).get("server", {}).get("algorithms", {}).get("cipher")
        del config_leftover["tailf-ned-cisco-ios-xr:ssh"]["server"]["algorithms"]["cipher"]

def main(before: dict, leftover: dict, translation_notes: list = []) -> dict:
    """
    Translates NSO Device configurations to MDD OpenConfig configurations.

    Requires environment variables:
    NSO_URL: str
    NSO_USERNAME: str
    NSO_PASSWORD: str
    NSO_DEVICE: str
    TEST - If True, sends generated OC configuration to NSO Server: str

    :param before: Original NSO Device configuration: dict
    :param leftover: NSO Device configuration minus configs replaced with MDD OC: dict
    :return: MDD Openconfig System configuration: dict
    """

    xr_system_config(before, leftover)
    xr_system_services(before, leftover)
    xr_system_ssh_server(before, leftover)
    translation_notes += system_notes

    return openconfig_system


if __name__ == "__main__":
    sys.path.append("../../")
    sys.path.append("../../../")

    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xr import common_xr
        from package_nso_to_oc import common
    else:
        import common_xr
        import common

    (config_before_dict, config_leftover_dict) = common_xr.init_xr_configs()
    main(config_before_dict, config_leftover_dict)
    config_name = "_system"
    config_remaining_name = "_remaining_system"
    oc_name = "_openconfig_system"
    common.print_and_test_configs("xr1", config_before_dict, config_leftover_dict, openconfig_system, 
        config_name, config_remaining_name, oc_name, system_notes)
else:
    # This is needed for now due to top level __init__.py. We need to determine if contents in __init__.py is still necessary.
    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xr import common_xr
        from package_nso_to_oc import common
    else:
        from xr import common_xr
        import common



#! /usr/bin/env python3
"""
Translate NSO Device config to MDD OpenConfig

This script will pull a device's configuration from an NSO server, convert the NED structured configuration to
MDD OpenConfig, save the NSO configuration to a file named {device_name}_ned_configuration_vlans.json, save the
NSO device configuration minus parts replaced by OpenConfig to a file named
{device_name}_ned_configuration_remaining_vlans.json, and save the MDD OpenConfig configuration to a file named
{nso_device}_openconfig_vlans.json.

The script requires the following environment variables:
NSO_URL - URL for the NSO server
NSO_USERNAME
NSO_PASSWORD
NSO_DEVICE - NSO device name for configuration translation
TEST - True or False. True enables sending the OpenConfig to the NSO server after generation
"""

import sys
from pathlib import Path
from importlib.util import find_spec

vlans_notes = []

openconfig_vlans = {
    "openconfig-network-instance:network-instances": {
        "openconfig-network-instance:network-instance": [
            {"openconfig-network-instance:name": "default",
             "openconfig-network-instance:config": {
                 "openconfig-network-instance:name": "default",
                 "openconfig-network-instance:type": "DEFAULT_INSTANCE",
                 "openconfig-network-instance:enabled": "true"
             },
             "openconfig-network-instance:vlans": {
                 "openconfig-network-instance:vlan": [
                 ]
             }
             }
        ]
    }
}


def xe_create_vlans(config_before: dict, config_leftover: dict) -> None:
    """
    Translates NSO XE NED to MDD OpenConfig Network Instances VLANs
    """
    openconfig_network_instance_vlans = \
        openconfig_vlans["openconfig-network-instance:network-instances"][
            "openconfig-network-instance:network-instance"][
            0]["openconfig-network-instance:vlans"]["openconfig-network-instance:vlan"]

    if config_before.get("tailf-ned-cisco-ios:vlan", {}).get("vlan-list"):
        for vlan in config_before["tailf-ned-cisco-ios:vlan"]["vlan-list"]:
            status = "ACTIVE"
            if vlan.get("shutdown"):
                status = "SUSPENDED"
            temp = {"openconfig-network-instance:vlan-id": vlan.get("id"),
                    "openconfig-network-instance:config": {
                        "openconfig-network-instance:vlan-id": vlan.get("id"),
                        "openconfig-network-instance:name": vlan.get("name", ""),
                        "openconfig-network-instance:status": status
                    }
                    }
            openconfig_network_instance_vlans.append(temp)
        del config_leftover["tailf-ned-cisco-ios:vlan"]


def main(before: dict, leftover: dict, translation_notes: list = []) -> dict:
    """
    Translates NSO Device configurations to MDD OpenConfig configurations.

    Requires environment variables:
    NSO_URL: str
    NSO_USERNAME: str
    NSO_PASSWORD: str
    NSO_DEVICE: str
    TEST - If True, sends generated OC configuration to NSO Server: str

    :param before: Original NSO Device configuration: dict
    :param leftover: NSO Device configuration minus configs replaced with MDD OC: dict
    :return: MDD Openconfig Network instances with VLANS configuration: dict
    """

    xe_create_vlans(before, leftover)
    translation_notes += vlans_notes

    return openconfig_vlans

if __name__ == "__main__":
    sys.path.append("../../")
    sys.path.append("../../../")

    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xe import common_xe
        from package_nso_to_oc import common
    else:
        import common_xe
        import common

    (config_before_dict, config_leftover_dict, interface_ip_dict) = common_xe.init_xe_configs()
    main(config_before_dict, config_leftover_dict)
    config_name = "_vlans"
    config_remaining_name = "_remaining_vlans"
    oc_name = "_openconfig_vlans"
    common.print_and_test_configs("xe1", config_before_dict, config_leftover_dict, openconfig_vlans, 
        config_name, config_remaining_name, oc_name, vlans_notes)
else:
    # This is needed for now due to top level __init__.py. We need to determine if contents in __init__.py is still necessary.
    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xe import common_xe
        from package_nso_to_oc import common
    else:
        from xe import common_xe
        import common



#! /usr/bin/env python3
"""
Translate NSO Device config to MDD OpenConfig

This script will pull a device's configuration from an NSO server, convert the NED structured configuration to
MDD OpenConfig, save the NSO configuration to a file named {device_name}_ned_configuration_network_instances.json,
save the NSO device configuration minus parts replaced by OpenConfig to a file named
{device_name}_ned_configuration_remaining_network_instances.json, and save the MDD OpenConfig configuration to a file
named {nso_device}_openconfig_acls.json.

The script requires the following environment variables:
NSO_URL - URL for the NSO server
NSO_USERNAME
NSO_PASSWORD
NSO_DEVICE - NSO device name for configuration translation
TEST - True or False. True enables sending the OpenConfig to the NSO server after generation
"""

import sys
from importlib.util import find_spec
from ipaddress import IPv4Network
import socket
import re

acls_notes = []
openconfig_acls = {
    "openconfig-acl:acl": {
        "openconfig-acl:acl-sets": {
            "openconfig-acl:acl-set": []
        },
        "openconfig-acl:interfaces": {
            "openconfig-acl:interface": []
        }
    }
}
protocols_oc_to_xr = {
    "icmp": "IP_ICMP",
    "igmp": "IP_IGMP",
    "ipnip": "IP_IN_IP",
    "tcp": "IP_TCP",
    "udp": "IP_UDP",
    "gre": "IP_GRE",
    "ahp": "IP_AUTH",
    "esp": "IP_ESP",
    "pim": "IP_PIM"
}
# OC has an additional forwarding action, "DROP", which also translates to "deny" in XR.
actions_xr_to_oc = {
    "permit": "ACCEPT",
    "deny": "REJECT",
}
port_operators = ["range", "eq", "lt", "gt", "neq"]
ACL_EXT_TYPE = "ACL_IPV4"


def xr_acls(config_before, config_after):
    oc_acl_set = openconfig_acls["openconfig-acl:acl"]["openconfig-acl:acl-sets"]["openconfig-acl:acl-set"]
    oc_acl_interface = openconfig_acls["openconfig-acl:acl"]["openconfig-acl:interfaces"]["openconfig-acl:interface"]
    access_list = config_before.get("tailf-ned-cisco-ios-xr:ipv4", {}).get("access-list", {})
    access_list_after = config_after.get("tailf-ned-cisco-ios-xr:ipv4", {}).get("access-list", {})
    interfaces_by_acl = get_interfaces_by_acl(config_before, config_after)
    acl_interfaces = {}

    for ext_index, ext_acl in enumerate(access_list.get("named-acl", [])):
        extended_acl = ExtendedAcl(oc_acl_set, ext_acl["rule"], ext_acl["name"],
                                   access_list_after["named-acl"][ext_index])
        extended_acl.process_acl()
        process_interfaces(ACL_EXT_TYPE, ext_acl["name"], interfaces_by_acl, acl_interfaces)

    for interface in acl_interfaces.values():
        oc_acl_interface.append(interface)

    process_ntp(config_before, config_after)
    process_line(config_before, config_after)


class BaseAcl:
    def __init__(self, oc_acl_set, xr_acl_set, xr_acl_name, xr_acl_set_after):
        self._oc_acl_set = oc_acl_set
        self._xr_acl_set = xr_acl_set
        self._xr_acl_set_after = xr_acl_set_after
        self._xr_acl_name = xr_acl_name
        self.acl_success = True

    def process_acl(self):
        acl_set = {
            "openconfig-acl:name": self._xr_acl_name,
            "openconfig-acl:type": self._acl_type,
            "openconfig-acl:config": {
                "openconfig-acl:name": self._xr_acl_name,
                "openconfig-acl:type": self._acl_type,
                "openconfig-acl:description": self._xr_acl_name,  # XR doesn't seem to have a description.
            },
            "openconfig-acl:acl-entries": {
                "openconfig-acl:acl-entry": []
            }
        }
        self.acl_success = True

        for rule_index, access_rule in enumerate(self._xr_acl_set):
            rule_success = self.__set_rule_parts(access_rule, acl_set)

            if rule_success:
                self._xr_acl_set_after['rule'][rule_index] = None
            else:
                self.acl_success = False

        # We only delete if all entries processed successfully.
        # We only add the ACL to OpenConfig if all entries processed successfully.
        if self.acl_success:
            self._oc_acl_set.append(acl_set)
            del self._xr_acl_set_after["name"]

    def __set_rule_parts(self, access_rule, acl_set):
        rule_parts = access_rule.get("line", "").split()

        if len(rule_parts) < 1:
            return

        success = True
        seq_id = access_rule.get("id")
        entry = {
            "openconfig-acl:sequence-id": seq_id,
            "openconfig-acl:config": {"openconfig-acl:sequence-id": seq_id},
            "openconfig-acl:actions": {
                "openconfig-acl:config": {"openconfig-acl:forwarding-action": actions_xr_to_oc[rule_parts[0]]}
            }
        }
        try:
            current_index = self.__set_protocol(entry, rule_parts)
            # Source IP
            current_index = self.__set_ip_and_port(rule_parts, current_index, entry, True)
            if self._acl_type == "ACL_IPV4":
                # Destination IP (if exists)
                current_index = self.__set_ip_and_port(rule_parts, current_index, entry, False)
        except Exception as err:
            success = False

        if (len(rule_parts) > current_index and rule_parts[current_index] == "log-input") or (
                len(rule_parts) > current_index and rule_parts[current_index] == "log"):
            entry["openconfig-acl:actions"]["openconfig-acl:config"]["openconfig-acl:log-action"] = "LOG_SYSLOG"
        else:
            entry["openconfig-acl:actions"]["openconfig-acl:config"]["openconfig-acl:log-action"] = "LOG_NONE"

        if success:
            acl_set["openconfig-acl:acl-entries"]["openconfig-acl:acl-entry"].append(entry)

        return success

    def __add_acl_entry_note(self, original_entry, note):
        acls_notes.append(f"""
            ACL name: {self._xr_acl_name}
            Original ACL entry: {original_entry}
            {note} 
        """)

    def __set_protocol(self, entry, rule_parts):
        if rule_parts[1] != 'ipv4':
            if not rule_parts[1] in protocols_oc_to_xr:
                self.__add_acl_entry_note(" ".join(rule_parts),
                                          f"protocol {rule_parts[1]} does not exist in expected list of protocols")
                self.acl_success = False
                raise ValueError
            self.__get_ipv4_config(entry)["openconfig-acl:protocol"] = protocols_oc_to_xr[rule_parts[1]]

        return 2

    def __get_ipv4_config(self, entry):
        if not self._ipv4_key in entry:
            entry[self._ipv4_key] = {}
        if not self._config_key in entry[self._ipv4_key]:
            entry[self._ipv4_key][self._config_key] = {}

        return entry[self._ipv4_key][self._config_key]

    def __get_transport_config(self, entry):
        if not "openconfig-acl:transport" in entry:
            entry["openconfig-acl:transport"] = {}
        if not "openconfig-acl:config" in entry["openconfig-acl:transport"]:
            entry["openconfig-acl:transport"]["openconfig-acl:config"] = {}

        return entry["openconfig-acl:transport"]["openconfig-acl:config"]

    def __set_ip_and_port(self, rule_parts, current_index, entry, is_source):
        if len(rule_parts) <= current_index:
            return current_index

        current_index = self.__set_ip_and_network(rule_parts, current_index, entry, is_source)

        if rule_parts[1] == "tcp" or rule_parts[1] == "udp":
            current_index = self.__set_port(rule_parts, current_index, entry, is_source)

        return current_index

    def __set_ip_and_network(self, rule_parts, current_index, entry, is_source):
        ip = rule_parts[current_index]

        if ip == "any":
            if is_source:
                self.__get_ipv4_config(entry)[self._src_addr_key] = "0.0.0.0/0"
            else:
                self.__get_ipv4_config(entry)["openconfig-acl:destination-address"] = "0.0.0.0/0"

            return current_index + 1
        elif ip == "host":
            if is_source:
                self.__get_ipv4_config(entry)[self._src_addr_key] = f"{rule_parts[current_index + 1]}/32"
            else:
                self.__get_ipv4_config(entry)[
                    "openconfig-acl:destination-address"] = f"{rule_parts[current_index + 1]}/32"

            return current_index + 2

        hostmask = rule_parts[current_index + 1]
        temp_ip = IPv4Network((0, hostmask))

        # 0.0.0.0 and 255.255.255.255 are wrong using IPv4Network.prefixlen()
        if hostmask == "0.0.0.0":
            prefixlen = "32"
        elif hostmask == "255.255.255.255":
            prefixlen = "0"
        else:
            prefixlen = temp_ip.prefixlen

        if is_source:
            self.__get_ipv4_config(entry)[self._src_addr_key] = f"{ip}/{prefixlen}"
        else:
            self.__get_ipv4_config(entry)["openconfig-acl:destination-address"] = f"{ip}/{prefixlen}"

        return current_index + 2

    def __set_port(self, rule_parts, current_index, entry, is_source):
        if len(rule_parts) <= current_index or not rule_parts[current_index] in port_operators:
            # We've either reached the end of the rule or there's no specified port
            if is_source:
                self.__get_transport_config(entry)["openconfig-acl:source-port"] = "ANY"
            else:
                self.__get_transport_config(entry)["openconfig-acl:destination-port"] = "ANY"
                current_index = self.__set_tcp_flags(rule_parts, current_index, entry)

            return current_index

        current_port = rule_parts[current_index + 1]

        try:
            current_port = current_port if current_port.isdigit() else socket.getservbyname(current_port)
        except OSError:
            try:
                current_port = common.port_name_number_mapping[current_port]
            except Exception as err:
                self.__add_acl_entry_note(" ".join(rule_parts),
                                          f"Unable to convert service {current_port} to a port number")
                self.acl_success = False
                raise Exception

        if rule_parts[current_index] == "range":
            end_port = rule_parts[current_index + 2]

            if is_source:
                self.__get_transport_config(entry)["openconfig-acl:source-port"] = f"{current_port}..{end_port}"
            else:
                self.__get_transport_config(entry)["openconfig-acl:destination-port"] = f"{current_port}..{end_port}"
                current_index = self.__set_tcp_flags(rule_parts, current_index + 3, entry)

            return current_index
        elif rule_parts[current_index] == "lt":
            if is_source:
                self.__get_transport_config(entry)["openconfig-acl:source-port"] = f"0..{int(current_port) - 1}"
            else:
                self.__get_transport_config(entry)["openconfig-acl:destination-port"] = f"0..{int(current_port) - 1}"
        elif rule_parts[current_index] == "gt":
            if is_source:
                self.__get_transport_config(entry)["openconfig-acl:source-port"] = f"{int(current_port) + 1}..65535"
            else:
                self.__get_transport_config(entry)[
                    "openconfig-acl:destination-port"] = f"{int(current_port) + 1}..65535"
        elif rule_parts[current_index] == "eq":
            if is_source:
                self.__get_transport_config(entry)["openconfig-acl:source-port"] = int(current_port)
            else:
                self.__get_transport_config(entry)["openconfig-acl:destination-port"] = int(current_port)
        elif rule_parts[current_index] == "neq":
            self.__add_acl_entry_note(" ".join(rule_parts),
                                      "XR ACL use of 'neq' port operator does not have an OC equivalent.")
            self.acl_success = False
            raise ValueError

        if not is_source:
            current_index = self.__set_tcp_flags(rule_parts, current_index + 2, entry)

        return current_index + 2

    def __set_tcp_flags(self, rule_parts, current_index, entry):
        if len(rule_parts) <= current_index or not rule_parts[current_index] in ["ack", "rst", "established"]:
            return current_index

        if rule_parts[current_index] == "ack":
            self.__get_transport_config(entry)["openconfig-acl:tcp-flags"] = ["TCP_ACK"]
        if rule_parts[current_index] == "rst":
            self.__get_transport_config(entry)["openconfig-acl:tcp-flags"] = ["TCP_RST"]
        if rule_parts[current_index] == "established":
            self.__get_transport_config(entry)["openconfig-acl:tcp-flags"] = ["TCP_ACK", "TCP_RST"]

        return current_index + 1


class ExtendedAcl(BaseAcl):
    def __init__(self, oc_acl_set, xr_acl_set, xr_acl_name, xr_acl_set_after):
        super(ExtendedAcl, self).__init__(oc_acl_set, xr_acl_set, xr_acl_name, xr_acl_set_after)
        self._rule_list_key = "named-acl"
        self._acl_type = "ACL_IPV4"
        self._ipv4_key = "openconfig-acl:ipv4"
        self._config_key = "openconfig-acl:config"
        self._src_addr_key = "openconfig-acl:source-address"


def get_interfaces_by_acl(config_before, config_after):
    interfaces_by_acl = {}
    interfaces = config_before.get("tailf-ned-cisco-ios-xr:interface", {})
    interfaces_after = config_after.get("tailf-ned-cisco-ios-xr:interface", {})
    for interface_type, interface_list in interfaces.items():
        interface_list_after = interfaces_after[interface_type]

        if "-subinterface" in interface_type:
            interface_list = interface_list[interface_type.replace("-subinterface", "")]
            interface_list_after = interface_list_after[interface_type.replace("-subinterface", "")]
            interface_type = interface_type.replace("-subinterface", "")
        if interface_type == "Bundle-Ether":
            interface_type = "Port-channel"

        for index, interface in enumerate(interface_list):
            if not "ipv4" in interface or not "access-group" in interface["ipv4"] or len(
                    interface["ipv4"]["access-group"]) < 1:
                continue

            intf_id = f"{interface_type}{interface['id']}"
            intf_numb_parts = re.split("[.]", interface["id"])
            intf_num = intf_numb_parts[0]
            subintf_num = int(intf_numb_parts[1]) if len(intf_numb_parts) > 1 else 0

            for access_group in interface["ipv4"]["access-group"]:
                if interface_list_after[index].get("ipv4") and interface_list_after[index]["ipv4"].get("access-group"):
                    del interface_list_after[index]["ipv4"]["access-group"]

                intf = {
                    "id": intf_id,
                    "interface": f"{interface_type}{intf_num}",
                    "subinterface": subintf_num,
                    "direction": access_group["direction"]
                }

                if not access_group["name"] in interfaces_by_acl:
                    interfaces_by_acl[access_group["name"]] = []

                interfaces_by_acl[access_group["name"]].append(intf)

    return interfaces_by_acl


def process_interfaces(acl_type, acl_name, interfaces_by_acl, acl_interfaces):
    interfaces = interfaces_by_acl.get(acl_name, [])

    for interface in interfaces:
        if interface["id"] in acl_interfaces:
            acl_interface = acl_interfaces[interface["id"]]
        else:
            acl_interface = {
                "openconfig-acl:id": interface["id"],
                "openconfig-acl:config": {"openconfig-acl:id": interface["id"]},
                "openconfig-acl:interface-ref": {
                    "openconfig-acl:config": {
                        "openconfig-acl:interface": interface["interface"],
                        "openconfig-acl:subinterface": interface["subinterface"]
                    }
                }
            }
            acl_interfaces[interface["id"]] = acl_interface

        intf_acl_set = get_intf_acl_set(acl_interface, interface["direction"])
        intf_acl_set.append({
            "openconfig-acl:set-name": acl_name,
            "openconfig-acl:type": acl_type,
            "openconfig-acl:config": {
                "openconfig-acl:set-name": acl_name,
                "openconfig-acl:type": acl_type
            }
        })


def get_intf_acl_set(acl_interface, direction):
    if direction == "ingress":
        ingress_set = "openconfig-acl:ingress-acl-set"
        if not f"{ingress_set}s" in acl_interface:
            acl_interface[f"{ingress_set}s"] = {ingress_set: []}
        if not ingress_set in acl_interface[f"{ingress_set}s"]:
            acl_interface[f"{ingress_set}s"][ingress_set] = []

        return acl_interface[f"{ingress_set}s"][ingress_set]
    elif direction == "egress":
        egress_set = "openconfig-acl:egress-acl-set"
        if not f"{egress_set}s" in acl_interface:
            acl_interface[f"{egress_set}s"] = {egress_set: []}
        if not egress_set in acl_interface[f"{egress_set}s"]:
            acl_interface[f"{egress_set}s"][egress_set] = []

        return acl_interface[f"{egress_set}s"][egress_set]
    else:
        raise ValueError("XR ACL not applied to interface with ingress or egress direction.")


def process_ntp(config_before, config_after):
    ntp_access_group = config_before.get("tailf-ned-cisco-ios-xr:ntp", {}).get("access-group", [])
    ntp_access_group_after = config_after.get("tailf-ned-cisco-ios-xr:ntp", {}).get("access-group", [])
    for acl_index, access_group in enumerate(ntp_access_group):

        if access_group.get("type") == "serve":
            ntp_peer = {
                "openconfig-acl-ext:server": {
                    "openconfig-acl-ext:config": {
                        "openconfig-acl-ext:server-acl-set": access_group.get("name")
                    }
                }
            }
            ntp_access_group_after[acl_index]["name"] = None
            ntp_access_group_after[acl_index]["type"] = None

        elif access_group.get("type") == "peer":
            ntp_peer = {
                "openconfig-acl-ext:peer": {
                    "openconfig-acl-ext:config": {
                        "openconfig-acl-ext:peer-acl-set": access_group.get("name")
                    }
                }
            }
            ntp_access_group_after[acl_index]["name"] = None
            ntp_access_group_after[acl_index]["type"] = None

        if ntp_peer:
            if openconfig_acls["openconfig-acl:acl"].get("openconfig-acl-ext:ntp"):
                openconfig_acls["openconfig-acl:acl"]["openconfig-acl-ext:ntp"].update(ntp_peer)
            else:
                openconfig_acls["openconfig-acl:acl"]["openconfig-acl-ext:ntp"] = ntp_peer


def check_default_profile_for_all_vty_lines(config_before) -> bool:
    """return True if default profile contains all vty-lines"""
    if not config_before.get("tailf-ned-cisco-ios-xr:vty-pool") or (
            len(config_before.get("tailf-ned-cisco-ios-xr:vty-pool")) == 1 and config_before.get(
        "tailf-ned-cisco-ios-xr:vty-pool").get("default")):
        return True
    elif config_before.get("tailf-ned-cisco-ios-xr:vty-pool", {}).get("default", {}).get(
            "first-vty") == 0 and config_before.get("tailf-ned-cisco-ios-xr:vty-pool", {}).get("default").get(
        "first-vty") == 99:
        return True
    else:
        return False


def acls_note_add(note):
    acls_notes.append(note)


def process_line(config_before, config_after):
    for name, line_item in config_before.get("tailf-ned-cisco-ios-xr:line", {}).items():
        if name == "default" and line_item.get("access-class"):
            openconfig_acls["openconfig-acl:acl"]["openconfig-acl-ext:lines"] = {"openconfig-acl-ext:line": []}
            acl_lines = openconfig_acls["openconfig-acl:acl"]["openconfig-acl-ext:lines"]["openconfig-acl-ext:line"]
            if not check_default_profile_for_all_vty_lines(config_before):
                access_group = f'{line_item.get("access-class").get("ingress")} {line_item.get("access-class").get("egress")}'
                acls_note_add(f"""
                    Could not apply VTY access group(s) {access_group}
                    MDD OpenConfig does not support multiple VTY line profiles.
                    Ensure all VTY lines (0 - 99) are in the 'default' VTY line profile.
                """)
            else:
                if line_item.get("access-class").get("ingress"):
                    acl_lines.append({
                        "openconfig-acl-ext:id": "vty 0 99",
                        "openconfig-acl-ext:config": {
                            "openconfig-acl-ext:id": "vty 0 99"},
                        "openconfig-acl-ext:ingress-acl-sets": {
                            "openconfig-acl-ext:ingress-acl-set": [
                                {"openconfig-acl-ext:ingress-acl-set-name": line_item.get("access-class").get(
                                    "ingress"),
                                    "openconfig-acl-ext:config": {
                                        "openconfig-acl-ext:ingress-acl-set-name": line_item.get("access-class").get(
                                            "ingress")}}]}})
                    config_after["tailf-ned-cisco-ios-xr:line"]["default"]["access-class"]["ingress"] = None
                if line_item.get("access-class").get("egress"):
                    acl_lines.append({
                        "openconfig-acl-ext:id": "vty 0 99",
                        "openconfig-acl-ext:config": {
                            "openconfig-acl-ext:id": "vty 0 99"},
                        "openconfig-acl-ext:egress-acl-set": line_item.get("access-class").get("egress")
                    })
                    config_after["tailf-ned-cisco-ios-xr:line"]["default"]["access-class"]["egress"] = None


def main(before: dict, leftover: dict, translation_notes: list = []) -> dict:
    """
    Translates NSO Device configurations to MDD OpenConfig configurations.

    Requires environment variables:
    NSO_URL: str
    NSO_USERNAME: str
    NSO_PASSWORD: str
    NSO_DEVICE: str
    TEST - If True, sends generated OC configuration to NSO Server: str

    :param before: Original NSO Device configuration: dict
    :param leftover: NSO Device configuration minus configs replaced with MDD OC: dict
    :param translation_notes: notes from previous NSO to OC translations if any
    :return: MDD Openconfig Network Instances configuration: dict
    """
    xr_acls(before, leftover)
    translation_notes += acls_notes

    return openconfig_acls


if __name__ == "__main__":
    sys.path.append("../../")
    sys.path.append("../../../")

    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xr import common_xr
        from package_nso_to_oc import common
    else:
        import common_xr
        import common

    (config_before_dict, config_leftover_dict) = common_xr.init_xr_configs()
    main(config_before_dict, config_leftover_dict)
    config_name = "_acls"
    config_remaining_name = "_remaining_acls"
    oc_name = "_openconfig_acls"
    common.print_and_test_configs(
        "xr1", config_before_dict, config_leftover_dict, openconfig_acls,
        config_name, config_remaining_name, oc_name, acls_notes)
else:
    # This is needed for now due to top level __init__.py. We need to determine if contents in __init__.py is still necessary.
    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xr import common_xr
        from package_nso_to_oc import common
    else:
        from xr import common_xr
        import common



#! /usr/bin/env python3
"""
This script is used by xe_network_instances.py to translate ospf configs from NED to OC.
"""

import copy
from functools import cmp_to_key
from importlib.util import find_spec

if (find_spec("package_nso_to_oc") is not None):
    from package_nso_to_oc.xe import common_xe
else:
    from xe import common_xe

ospf_network_types = {
    "broadcast": "BROADCAST_NETWORK",
    "point-to-point": "POINT_TO_POINT_NETWORK",
    "non-broadcast": "NON_BROADCAST_NETWORK"
}
xe_ospf_notes = []

def configure_xe_ospf(net_inst, vrf_interfaces, config_before, config_leftover, network_instances_notes):
    """
    Translates NSO XE NED to MDD OpenConfig Network Instances
    """
    instance_type = net_inst["openconfig-network-instance:config"]["openconfig-network-instance:type"]
    net_protocols = net_inst["openconfig-network-instance:protocols"]["openconfig-network-instance:protocol"]
    ospf_list = config_before.get("tailf-ned-cisco-ios:router", {}).get("ospf")

    if ospf_list == None:
        return

    for ospf_index, ospf in enumerate(ospf_list):
        if ((instance_type == "L3VRF" and "vrf" in ospf)
                or (instance_type == "DEFAULT_INSTANCE" and not "vrf" in ospf)):
            process_ospf(net_protocols, vrf_interfaces, config_leftover, ospf_index, ospf)

    network_instances_notes += xe_ospf_notes


def configure_xe_ospf_redistribution(net_inst, config_before, config_leftover, router_ospf_by_vrf):
    ospf_before = config_before.get("tailf-ned-cisco-ios:router", {"ospf": []}).get("ospf")

    if ospf_before == None or len(ospf_before) == 0:
        return
        
    instance_name = net_inst["openconfig-network-instance:name"]
    ospf_leftover = config_leftover.get("tailf-ned-cisco-ios:router", {"ospf": []}).get("ospf")

    for router_ospf_index in router_ospf_by_vrf.get(instance_name, []):
        router_ospf_before = ospf_before[router_ospf_index]
        router_ospf_leftover = ospf_leftover[router_ospf_index]
        redistribute = router_ospf_before.get("redistribute", {})
        redistribute_leftover = router_ospf_leftover.get("redistribute", {})

        if len(redistribute) == 0:
            continue

        common_xe.process_redistribute(net_inst, redistribute, redistribute_leftover, "OSPF", 
            router_ospf_before["id"])

        if "redistribute" in router_ospf_leftover and len(redistribute_leftover) == 0:
            del router_ospf_leftover["redistribute"]
        if "vrf" in router_ospf_leftover:
            del router_ospf_leftover["vrf"]
        if len(router_ospf_leftover) == 1 and "id" in router_ospf_leftover:
            del router_ospf_leftover["id"]


def get_interfaces_by_area(ospf_id, network_statements, vrf_interfaces):
    """
    Assigns OSPF enabled interfaces by area, based on OSPF network statements.
    Network statement wildcard masks are treated like ACLs to determine which interface will be attached
    to an OSPF area.
    Source: https://www.cisco.com/c/en/us/td/docs/ios-xml/ios/iproute_ospf/command/iro-cr-book/ospf-i1.html#wp2261032279

    TODO: Consider creating network statement sort algorithm
    As there can be multiple network statements, some may be more specific than others and they can point to different
    OSPF areas. The more specific statements should be prioritized over the less specific ones.
    """
    processed_interfaces = set()
    interfaces_by_area = {}
    sorted_network_statements = sorted(network_statements, key=cmp_to_key(sort_by_mask))
    unmatched_statements = set()
    matched_statements_with_intf = set()
    # This is for tracking only, because the same statement can appear again for unmatched statements.
    matched_statements = set()

    if len(sorted_network_statements) > 0:
        xe_ospf_notes.append("\n\nNo direct translation for network statements from XE to OC. \
Below are the leftover network statements.")
        xe_ospf_notes.append(f"OSPF ID: {ospf_id}")

    for net_stmt in sorted_network_statements:
        # Use get method for net_stmt, since it contains input values that we did not generate and cannot guarantee.
        stmt_mask = net_stmt.get("mask", "")
        area_id = net_stmt.get("area", "0")
        merged_statement_ip = binary_merge(net_stmt.get("ip", ""), net_stmt.get("mask", ""))

        for vrf_intf in vrf_interfaces:
            full_intf_name = vrf_intf["type"] + vrf_intf["name"]
            intf_addr = vrf_intf["ip"]["address"]["primary"]["address"]
            merged_vrf_intf_ip = binary_merge(intf_addr, stmt_mask)

            if merged_statement_ip == merged_vrf_intf_ip:
                matched_stmt = {
                    "interface": full_intf_name,
                    "interface_ip": intf_addr
                }
                matched_stmt.update(net_stmt)
                matched_statements.add(net_stmt_to_str(net_stmt))
                matched_statements_with_intf.add(net_stmt_to_str(matched_stmt))

                if (full_intf_name in processed_interfaces):
                    continue

                # If there's a match, then this interface is OSPF enabled
                processed_interfaces.add(full_intf_name)

                if not area_id in interfaces_by_area:
                    interfaces_by_area[area_id] = []

                interfaces_by_area[area_id].append(vrf_intf)
            else:
                unmatched_statements.add(net_stmt_to_str(net_stmt))

    if len(unmatched_statements) > 0 or len(matched_statements_with_intf) > 0:
        unmatched_statements -= matched_statements
        xe_ospf_notes.append("Matched statements:")
        for matched_statement in matched_statements_with_intf: xe_ospf_notes.append(matched_statement)
        xe_ospf_notes.append("Non-matching statements:")
        for unmatched_statement in unmatched_statements: xe_ospf_notes.append(unmatched_statement)
        xe_ospf_notes.append("\n")

    return interfaces_by_area


def net_stmt_to_str(net_stmt):
    net_stmt_str = ""
    
    if "interface" in net_stmt:
        net_stmt_str += f"\tInterface: {net_stmt['interface']}, Interface IP: {net_stmt['interface_ip']}\n"
    
    net_stmt_str += f"\tMask: {net_stmt.get('mask', '')}, IP: {net_stmt.get('ip', '')}, Area: {net_stmt.get('area', '')}"

    return net_stmt_str


def sort_by_mask(stmt1, stmt2):
    """
    A comparator to sort by mask, ordered from most specific (mask of 0.0.0.0) to least specific (mask of 255.255.255.255).
    """
    mask1 = stmt1.get("mask", None)
    mask2 = stmt2.get("mask", None)

    if not mask1 and not mask2:
        return 0
    if not mask1:
        return 1
    if not mask2:
        return -1

    mask1_octets = mask1.split(".")
    mask2_octets = mask2.split(".")

    # Sanity check
    if len(mask1_octets) != 4 or len(mask2_octets) != 4:
        raise ValueError("Invalid IP string provided")

    for octet_index in range(len(mask1_octets)):
        mask1_octet_int = int(mask1_octets[octet_index])
        mask2_octet_int = int(mask2_octets[octet_index])

        if mask1_octet_int == mask2_octet_int:
            continue

        return mask1_octet_int - mask2_octet_int

    return 0


def binary_merge(ip, mask):
    """
    Transform the IPs into binary string format and merge the binary strings via OR operation
    """
    merged_result = []
    ip_in_binary = get_binary_str(ip)
    mask_in_binary = get_binary_str(mask)

    for index in range(len(ip_in_binary)):
        ip_val = ip_in_binary[index] == "1"
        mask_val = mask_in_binary[index] == "1"
        merged_result.append("1" if ip_val or mask_val else "0")

    return "".join(merged_result)


def get_binary_str(ip_str):
    octets = ip_str.split(".")
    binary_octets = []

    for octet in octets:
        binary_octets.append(format(int(octet), "08b"))

    return "".join(binary_octets)


def get_ospfv2_global(net_protocols, prot_index):
    if (len(net_protocols) >= prot_index):
        if not "openconfig-network-instance:ospfv2" in net_protocols[prot_index]:
            net_protocols[prot_index]["openconfig-network-instance:ospfv2"] = {}
        if not "openconfig-network-instance:global" in net_protocols[prot_index]["openconfig-network-instance:ospfv2"]:
            net_protocols[prot_index]["openconfig-network-instance:ospfv2"]["openconfig-network-instance:global"] = {}

        return net_protocols[prot_index]["openconfig-network-instance:ospfv2"]["openconfig-network-instance:global"]
    else:
        # Sanity check, should not occur...
        raise IndexError(f"The protocol index {prot_index} does not exist.")


def get_ospfv2_area(net_protocols, prot_index):
    if (len(net_protocols) >= prot_index):
        if not "openconfig-network-instance:ospfv2" in net_protocols[prot_index]:
            net_protocols[prot_index]["openconfig-network-instance:ospfv2"] = {}
        if not "areas" in net_protocols[prot_index]["openconfig-network-instance:ospfv2"]:
            net_protocols[prot_index]["openconfig-network-instance:ospfv2"]["openconfig-network-instance:areas"] = {
                "openconfig-network-instance:area": []}

        return net_protocols[prot_index]["openconfig-network-instance:ospfv2"]["openconfig-network-instance:areas"][
            "openconfig-network-instance:area"]
    else:
        # Sanity check, should not occur...
        raise IndexError(f"The protocol index {prot_index} does not exist.")


def get_area_by_id(ospfv2_area, area_id):
    for area in ospfv2_area:
        if area["openconfig-network-instance:identifier"] == area_id:
            return area

    new_area = {"openconfig-network-instance:identifier": area_id,
                "openconfig-network-instance:config": {"openconfig-network-instance:identifier": area_id}}
    ospfv2_area.append(new_area)

    return new_area


def is_area_present_by_id(ospfv2_area, id):
    for area in ospfv2_area:
        if area.get("openconfig-network-instance:identifier", None) == id:
            return True

    return False


def get_intf_by_intf_number(intf_attr, intf_number):
    for intf in intf_attr:
        if str(intf["name"]) == intf_number:
            return intf


def process_ospf(net_protocols, vrf_interfaces, config_leftover, ospf_index, ospf):
    ospf_leftover = config_leftover.get("tailf-ned-cisco-ios:router", {}).get("ospf")[ospf_index]
    # If we got here, we init an empty dict and append to protocol list for future use.
    net_protocols.append({})
    prot_index = len(net_protocols) - 1
    set_network_config(ospf_leftover, net_protocols, prot_index, ospf)
    set_ospf2_global_config(ospf_leftover, net_protocols, prot_index, ospf)
    set_graceful_restart_ietf(ospf_leftover, net_protocols, prot_index, ospf)
    set_vrf_lite(ospf_leftover, net_protocols, prot_index, ospf)
    set_default_info_originate(ospf_leftover, net_protocols, prot_index, ospf)
    check_areas(ospf_leftover, net_protocols, vrf_interfaces, config_leftover, prot_index, ospf)
    set_mpls_ldp_sync(ospf_leftover, net_protocols, prot_index, ospf)
    set_timers_lsa(ospf_leftover, net_protocols, prot_index, ospf)
    set_timers_spf(ospf_leftover, net_protocols, prot_index, ospf)
    set_auto_cost_ref_bandwidth(ospf_leftover, net_protocols, prot_index, ospf)


def set_network_config(ospf_leftover, net_protocols, prot_index, ospf):
    net_protocols[prot_index]["openconfig-network-instance:identifier"] = "OSPF"
    net_protocols[prot_index]["openconfig-network-instance:name"] = f'{ospf.get("id")}'
    temp_ospf = {"openconfig-network-instance:config": {
        "openconfig-network-instance:identifier": "OSPF",
        "openconfig-network-instance:name": f'{ospf.get("id")}',
        "openconfig-network-instance:enabled": True
    }}
    net_protocols[prot_index].update(temp_ospf)

    # Don't remove keys until everything is in OC
    # if ospf_leftover.get("id"):
    #     del ospf_leftover["id"]


def set_ospf2_global_config(ospf_leftover, net_protocols, prot_index, ospf):
    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)

    if (not ospf.get("router-id") and not ospf.get("log-adjacency-changes")
            and not ospf.get("compatible") and not ospf.get("prefix-suppression")):
        ospfv2_global["config"] = {
            "openconfig-network-instance:log-adjacency-changes": False,
            "openconfig-network-instance:summary-route-cost-mode": "RFC1583_COMPATIBLE",
            "openconfig-network-instance:hide-transit-only-networks": False
        }

        return

    ospfv2_global_config = {}

    if ospf.get("router-id"):
        ospfv2_global_config["openconfig-network-instance:router-id"] = f'{ospf.get("router-id")}'
        del ospf_leftover["router-id"]
    if ospf.get("log-adjacency-changes"):
        ospfv2_global_config["openconfig-network-instance:log-adjacency-changes"] = True
    else:
        ospfv2_global_config["openconfig-network-instance:log-adjacency-changes"] = False
    if ospf.get("compatible") and ospf["compatible"].get("rfc1583") is False:
        ospfv2_global_config["openconfig-network-instance:summary-route-cost-mode"] = "RFC2328_COMPATIBLE"
    else:
        ospfv2_global_config["openconfig-network-instance:summary-route-cost-mode"] = "RFC1583_COMPATIBLE"
    if ospf.get("prefix-suppression"):
        ospfv2_global_config["openconfig-network-instance:hide-transit-only-networks"] = True
    else:
        ospfv2_global_config["openconfig-network-instance:hide-transit-only-networks"] = False

    # Common clean up
    if "log-adjacency-changes" in ospf_leftover:
        del ospf_leftover["log-adjacency-changes"]
    if "compatible" in ospf_leftover:
        del ospf_leftover["compatible"]
    if "prefix-suppression" in ospf_leftover:
        del ospf_leftover["prefix-suppression"]

    ospfv2_global["openconfig-network-instance:config"] = ospfv2_global_config


def set_graceful_restart_ietf(ospf_leftover, net_protocols, prot_index, ospf):
    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)
    graceful_restart_config = {
        "openconfig-network-instance:graceful-restart": {"openconfig-network-instance:config": {}}}

    if ospf.get("nsf-ietf") and ospf["nsf-ietf"].get("nsf") and "ietf" in ospf["nsf-ietf"]["nsf"]:
        graceful_restart_config["openconfig-network-instance:graceful-restart"]["openconfig-network-instance:config"][
            "openconfig-network-instance:enabled"] = True
        del ospf_leftover["nsf-ietf"]["nsf"]["ietf"]
    else:
        graceful_restart_config["openconfig-network-instance:graceful-restart"]["openconfig-network-instance:config"][
            "openconfig-network-instance:enabled"] = False

    ospfv2_global.update(graceful_restart_config)


def set_vrf_lite(ospf_leftover, net_protocols, prot_index, ospf):
    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)

    if not ospfv2_global.get("openconfig-network-instance:config"):
        ospfv2_global["openconfig-network-instance:config"] = {}

    if ospf.get("capability") and "vrf-lite" in ospf["capability"]:
        ospfv2_global["openconfig-network-instance:config"]["openconfig-ospfv2-ext:capability-vrf-lite"] = True
        del ospf_leftover["capability"]
    else:
        ospfv2_global["openconfig-network-instance:config"]["openconfig-ospfv2-ext:capability-vrf-lite"] = False


def set_default_info_originate(ospf_leftover, net_protocols, prot_index, ospf):
    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)

    if not "default-information" in ospf or not "originate" in ospf["default-information"]:
        ospfv2_global["openconfig-network-instance:config"].update({
                                                                       "openconfig-ospfv2-ext:default-information-originate": {
                                                                           "openconfig-ospfv2-ext:config": {
                                                                               "openconfig-ospfv2-ext:enabled": False}}})
        return

    if not "openconfig-network-instance:config" in ospfv2_global:
        ospfv2_global["openconfig-network-instance:config"] = {}

    oc_default_info_originate = {"openconfig-ospfv2-ext:default-information-originate": {
        "openconfig-ospfv2-ext:config": {"openconfig-ospfv2-ext:enabled": True}}}
    originate_config = oc_default_info_originate["openconfig-ospfv2-ext:default-information-originate"][
        "openconfig-ospfv2-ext:config"]
    default_info_originate = ospf["default-information"]["originate"]

    if "always" in default_info_originate:
        originate_config["openconfig-ospfv2-ext:always"] = True

        if "always" in ospf_leftover["default-information"]["originate"]:
            del ospf_leftover["default-information"]["originate"]["always"]
    if "metric" in default_info_originate:
        originate_config["openconfig-ospfv2-ext:metric"] = default_info_originate["metric"]

        if "metric" in ospf_leftover["default-information"]["originate"]:
            del ospf_leftover["default-information"]["originate"]["metric"]
    if "metric-type" in default_info_originate:
        originate_config["openconfig-ospfv2-ext:metric-type"] = default_info_originate["metric-type"]

        if "metric-type" in ospf_leftover["default-information"]["originate"]:
            del ospf_leftover["default-information"]["originate"]["metric-type"]
    # TODO add route-maps to OC Services
    # if "route-map" in default_info_originate:
    #     originate_config["route-map"] = default_info_originate["route-map"]
    #
    #     if "route-map" in ospf_leftover["default-information"]["originate"]:
    #         del ospf_leftover["default-information"]["originate"]["route-map"]

    ospfv2_global["openconfig-network-instance:config"].update(oc_default_info_originate)


def check_areas(ospf_leftover, net_protocols, vrf_interfaces, config_leftover, prot_index, ospf):
    intf_config_leftover = config_leftover.get("tailf-ned-cisco-ios:interface", {})
    ospfv2_area = get_ospfv2_area(net_protocols, prot_index)
    interfaces_by_area = get_interfaces_by_area(ospf.get("id"), ospf.get("network", []), vrf_interfaces)
    area_list = populate_area_list(ospf)
    is_area_0_present = check_for_area_0(area_list)
    for area in area_list:
        leftover_area = list(filter(lambda temp_area: temp_area["id"] == area["id"], ospf_leftover.get("area", [])))
        set_ospfv2_areas(ospfv2_area, area, leftover_area, ospf, ospf_leftover)

        if is_area_0_present and int(area["id"]) != 0:
            # We do this as long as area 0 is available and destination area is not 0.
            set_inter_area_propagation_policy(net_protocols, prot_index, area, leftover_area)

        for current_intf in interfaces_by_area.get(area.get("id", 0), []):
            intf_type, intf_number = (current_intf["type"], current_intf["name"])
            intf_name = intf_type + intf_number
            if (intf_type == "Port-channel" or intf_type == "LISP") and "." in str(intf_number):
                intf_attr_leftover = intf_config_leftover.get(f"{intf_type}-subinterface", {}).get(intf_type, {})
            else:
                intf_attr_leftover = intf_config_leftover.get(intf_type, {})
            intf_leftover = get_intf_by_intf_number(intf_attr_leftover, intf_number)
            set_ospfv2_intf_areas(ospfv2_area, intf_leftover, area, intf_name, current_intf, ospf, ospf_leftover)


def populate_area_list(ospf):
    area_list = []
    area_id_set = set()

    # Populate with existing areas defined by NED first, if any.
    for area in ospf.get("area", []):
        area_list.append(area)
        area_id_set.add(area["id"])

    # Populate with area based on OSPF network statement, if area wasn't already defined by the NED
    for net_stmt in ospf.get("network", []):
        # Is adding a default necessary, or should we always expect a network statement to contain an area ID?
        area_id = net_stmt.get("area", "0")

        if area_id in area_id_set:
            continue
        else:
            area_id_set.add(area_id)

        area_list.append({"id": area_id})
    return area_list


def check_for_area_0(area_list):
    for area in area_list:
        if area["id"] == 0:
            return True

    return False


def set_inter_area_propagation_policy(net_protocols, prot_index, area, leftover_area):
    if "id" in area and "filter-list" in area and len(area["filter-list"]) == 1:
        ospfv2_global = get_ospfv2_global(net_protocols, prot_index)
        import_policy_prefix = area["filter-list"][0]["prefix"]
        service_policy = {}

        if not "openconfig-network-instance:inter-area-propagation-policies" in ospfv2_global:
            ospfv2_global["openconfig-network-instance:inter-area-propagation-policies"] = {}
        if not "openconfig-network-instance:inter-area-propagation-policy" in ospfv2_global[
            "openconfig-network-instance:inter-area-propagation-policies"]:
            ospfv2_global["openconfig-network-instance:inter-area-propagation-policies"][
                "openconfig-network-instance:inter-area-propagation-policy"] = []

        # Per Steven Mosher, 0 is the implied source area
        service_policy["openconfig-network-instance:src-area"] = 0
        service_policy["openconfig-network-instance:dst-area"] = area["id"]
        service_policy["openconfig-network-instance:config"] = {
            "openconfig-network-instance:src-area": 0,
            "openconfig-network-instance:dst-area": area["id"],
            "openconfig-network-instance:import-policy": [import_policy_prefix]
        }

        ospfv2_global["openconfig-network-instance:inter-area-propagation-policies"][
            "openconfig-network-instance:inter-area-propagation-policy"].append(service_policy)

        if len(leftover_area) > 0 and "filter-list" in leftover_area[0]:
            del leftover_area[0]["filter-list"]


def set_mpls_ldp_sync(ospf_leftover, net_protocols, prot_index, ospf):
    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)
    is_igp_ldp_sync = "mpls" in ospf and "ldp" in ospf["mpls"] and "sync" in ospf["mpls"]["ldp"]
    ospfv2_global["openconfig-network-instance:mpls"] = {"openconfig-network-instance:igp-ldp-sync": {
        "openconfig-network-instance:config": {"openconfig-network-instance:enabled": is_igp_ldp_sync}}}

    if is_igp_ldp_sync:
        del ospf_leftover["mpls"]["ldp"]["sync"]


def set_timers_lsa(ospf_leftover, net_protocols, prot_index, ospf):
    if not "timers" in ospf or not "throttle" in ospf["timers"] or not "lsa" in ospf["timers"]["throttle"]:
        return

    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)
    config = {}
    lsa = ospf["timers"]["throttle"]["lsa"]

    if "start-interval" in lsa or "hold-interval" in lsa or "max-interval" in lsa:
        if "start-interval" in lsa and "hold-interval" in lsa and "max-interval" in lsa:
            config["openconfig-network-instance:initial-delay"] = lsa["start-interval"]
            config["openconfig-network-instance:maximum-delay"] = lsa["max-interval"]
            config["openconfig-ospfv2-ext:hold-time"] = lsa["hold-interval"]

            if not "openconfig-network-instance:timers" in ospfv2_global:
                ospfv2_global["openconfig-network-instance:timers"] = {}
            if not "openconfig-network-instance:lsa-generation" in ospfv2_global["openconfig-network-instance:timers"]:
                ospfv2_global["openconfig-network-instance:timers"]["openconfig-network-instance:lsa-generation"] = {}

            ospfv2_global["openconfig-network-instance:timers"]["openconfig-network-instance:lsa-generation"].update(
                {"openconfig-network-instance:config": config})
        else:
            raise ValueError(
                "XE OSPF throttle timers lsa needs values for start-interval, hold-interval, and max-interval")

    del ospf_leftover["timers"]["throttle"]["lsa"]


def set_timers_spf(ospf_leftover, net_protocols, prot_index, ospf):
    if not "timers" in ospf or not "throttle" in ospf["timers"] or not "spf" in ospf["timers"]["throttle"]:
        return

    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)
    config = {}
    spf = ospf["timers"]["throttle"]["spf"]

    if "spf-start" in spf or "spf-hold" in spf or "spf-max-wait" in spf:
        if "spf-start" in spf and "spf-hold" in spf and "spf-max-wait" in spf:
            config["openconfig-network-instance:initial-delay"] = spf["spf-start"]
            config["openconfig-network-instance:maximum-delay"] = spf["spf-max-wait"]
            config["openconfig-ospfv2-ext:hold-time"] = spf["spf-hold"]

            if not "openconfig-network-instance:timers" in ospfv2_global:
                ospfv2_global["openconfig-network-instance:timers"] = {}
            if not "openconfig-network-instance:spf" in ospfv2_global["openconfig-network-instance:timers"]:
                ospfv2_global["openconfig-network-instance:timers"]["openconfig-network-instance:spf"] = {}

            ospfv2_global["openconfig-network-instance:timers"]["openconfig-network-instance:spf"].update(
                {"openconfig-network-instance:config": config})
        else:
            raise ValueError("XE OSPF throttle timers spf needs values for spf-start, spf-hold, and spf-max-wait")

    del ospf_leftover["timers"]["throttle"]["spf"]


def set_auto_cost_ref_bandwidth(ospf_leftover, net_protocols, prot_index, ospf):
    # if not "auto-cost reference-bandwidth" in ospf:
    if not "auto-cost" in ospf or not "reference-bandwidth" in ospf["auto-cost"]:
        return

    ospfv2_global = get_ospfv2_global(net_protocols, prot_index)
    auto_cost = ospf["auto-cost"]

    ospfv2_global["openconfig-network-instance:config"].update(
        {"openconfig-ospfv2-ext:auto-cost-ref-bandwidth": auto_cost["reference-bandwidth"]})

    del ospf_leftover["auto-cost"]["reference-bandwidth"]


def set_ospfv2_intf_areas(ospfv2_area, intf_leftover, area, intf_name, intf, ospf, ospf_leftover):
    intf_config = {"openconfig-network-instance:id": intf_name}
    set_network_type(intf, intf_leftover, intf_config)
    set_metric(intf, intf_leftover, intf_config)
    set_passive(ospf, ospf_leftover, intf_config, intf_name)
    set_priority(intf, intf_leftover, intf_config)
    area_intf = get_area_by_id(ospfv2_area, area["id"])

    if not "openconfig-network-instance:interfaces" in area_intf:
        area_intf["openconfig-network-instance:interfaces"] = {"openconfig-network-instance:interface": []}
    if not "openconfig-network-instance:interface" in area_intf["openconfig-network-instance:interfaces"]:
        area_intf["openconfig-network-instance:interfaces"]["openconfig-network-instance:interface"] = []

    area_intf["openconfig-network-instance:interfaces"]["openconfig-network-instance:interface"].append({
        "openconfig-network-instance:id": intf_name,
        "openconfig-network-instance:config": intf_config,
        "openconfig-network-instance:enable-bfd": {"openconfig-network-instance:config": {
            "openconfig-network-instance:enabled": is_bfd_enabled(intf, intf_leftover)}},
        "openconfig-network-instance:neighbors": set_neighbors(ospf, ospf_leftover),
        "openconfig-network-instance:timers": set_timers(intf, intf_leftover),
        "openconfig-ospfv2-ext:authentication": set_authentication(intf, intf_leftover),
    })


def set_network_type(intf, intf_leftover, intf_config):
    if ("ip" in intf and "ospf" in intf["ip"] and "network" in intf["ip"]["ospf"] and len(
            intf["ip"]["ospf"]["network"]) > 0
            and intf["ip"]["ospf"]["network"][0] in ospf_network_types):
        intf_config["openconfig-network-instance:network-type"] = ospf_network_types[intf["ip"]["ospf"]["network"][0]]

        if "network" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["network"]


def set_metric(intf, intf_leftover, intf_config):
    if "ip" in intf and "ospf" in intf["ip"] and "cost" in intf["ip"]["ospf"]:
        intf_config["openconfig-network-instance:metric"] = intf["ip"]["ospf"]["cost"]

        if "cost" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["cost"]


def set_passive(ospf, ospf_leftover, intf_config, intf_name):
    if "passive-interface" in ospf and "interface" in ospf["passive-interface"]:
        # We're brute forcing, but we don't expect 100s of interfaces anyway...
        for passive_intf in ospf["passive-interface"]["interface"]:
            if passive_intf["name"] == intf_name:
                intf_config["openconfig-network-instance:passive"] = True
                break
        else:
            intf_config["passive"] = False

    if ("passive-interface" in ospf_leftover):
        del ospf_leftover["passive-interface"]


def set_priority(intf, intf_leftover, intf_config):
    if "ip" in intf and "ospf" in intf["ip"] and "priority" in intf["ip"]["ospf"]:
        intf_config["openconfig-network-instance:priority"] = intf["ip"]["ospf"]["priority"]

        if "priority" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["priority"]


def is_bfd_enabled(intf, intf_leftover):
    if "ip" in intf and "ospf" in intf["ip"] and "bfd" in intf["ip"]["ospf"]:
        if "bfd" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["bfd"]

        return True

    return False


def set_neighbors(ospf, ospf_leftover):
    neighbor = []
    if "neighbor" in ospf:
        for neighbor_index, ospf_neighbor in enumerate(ospf["neighbor"]):
            metric = ""

            if "cost-database-filter-container" in ospf_neighbor and "cost" in ospf_neighbor[
                "cost-database-filter-container"]:
                metric = ospf_neighbor["cost-database-filter-container"]["cost"]

            neighbor.append({
                "openconfig-network-instance:router-id": ospf_neighbor.get("ip", ""),
                "openconfig-network-instance:config": {
                    "openconfig-network-instance:router-id": ospf_neighbor.get("ip", ""),
                    "openconfig-network-instance:metric": metric
                }
            })

            # Deleting indexes while iterating causes issues, just blank out or nullify
            ospf_leftover["neighbor"][neighbor_index] = None

    return {"neighbor": neighbor}


def set_timers(intf, intf_leftover):
    config = {}

    if "ip" in intf and "ospf" in intf["ip"] and "hello-interval" in intf["ip"]["ospf"]:
        config["openconfig-network-instance:hello-interval"] = intf["ip"]["ospf"]["hello-interval"]

        if "hello-interval" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["hello-interval"]
    if "ip" in intf and "ospf" in intf["ip"] and "retransmit-interval" in intf["ip"]["ospf"]:
        config["openconfig-network-instance:retransmission-interval"] = intf["ip"]["ospf"]["retransmit-interval"]

        if "retransmit-interval" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["retransmit-interval"]
    if ("ip" in intf and "ospf" in intf["ip"] and "dead-interval" in intf["ip"]["ospf"]
            and "seconds" in intf["ip"]["ospf"]["dead-interval"]):
        config["openconfig-network-instance:dead-interval"] = intf["ip"]["ospf"]["dead-interval"]["seconds"]

        if "dead-interval" in intf_leftover["ip"]["ospf"]:
            del intf_leftover["ip"]["ospf"]["dead-interval"]

    return {"openconfig-network-instance:config": config}


def set_authentication(intf, intf_leftover):
    # Authentication type: unconfigured, null, simple and md5
    authentication = {"openconfig-ospfv2-ext:config": []}
    auth_list = authentication["openconfig-ospfv2-ext:config"]
    is_auth_enabled = type(intf.get("ip", {}).get("ospf", {}).get("authentication", '')) is dict
    is_mess_digest = "ip" in intf and "ospf" in intf["ip"] and "message-digest-key" in intf["ip"]["ospf"]
    if not is_auth_enabled:
        # Unconfigured
        config = {
            "openconfig-ospfv2-ext:authentication-type": 'UNCONFIGURED'
        }
        auth_list.append(config)
    elif is_auth_enabled and len(intf["ip"]["ospf"]["authentication"]) == 0:
        # Simple
        config = {
            "openconfig-ospfv2-ext:authentication-type": "SIMPLE",
        }
        del intf_leftover["ip"]["ospf"]["authentication"]
        if intf.get("ip", {}).get("ospf", {}).get("authentication-key", {}).get("secret", ""):
            config["openconfig-ospfv2-ext:simple-password"] = intf.get("ip", {}).get("ospf", {}).get(
                "authentication-key", {}).get("secret", "")
            intf["ip"]["ospf"]["authentication-key"]["secret"] = None
        auth_list.append(config)
    else:
        updated_ospf_auth_list = []

        for index, auth in enumerate(intf["ip"]["ospf"]["authentication"]):
            # NULL
            if "null" in intf["ip"]["ospf"]["authentication"]:
                config = {
                    "openconfig-ospfv2-ext:authentication-type": 'NULL'
                }
                intf_leftover["ip"]["ospf"]["authentication"][index] = None
                auth_list.append(config)
            # MD5
            if "message-digest" in intf["ip"]["ospf"]["authentication"]:
                config = {
                    "openconfig-ospfv2-ext:authentication-type": 'MD5'
                }
                intf_leftover["ip"]["ospf"]["authentication"][index] = None
                auth_list.append(config)
            if intf_leftover["ip"]["ospf"]["authentication"][index]:
                updated_ospf_auth_list.append(intf_leftover["ip"]["ospf"]["authentication"][index])

        if len(updated_ospf_auth_list) > 0:
            intf_leftover["ip"]["ospf"]["authentication"] = updated_ospf_auth_list
        else:
            del intf_leftover["ip"]["ospf"]["authentication"]

    if is_mess_digest:
        authentication.update(set_message_digest(intf, intf_leftover))

    return authentication


def set_message_digest(intf, intf_leftover):
    # Configure md5 keys
    mess_digest = {"openconfig-ospfv2-ext:md5-authentication-keys": {
        "openconfig-ospfv2-ext:md5-authentication-key": []
    }}
    mess_digest_list = mess_digest["openconfig-ospfv2-ext:md5-authentication-keys"][
        "openconfig-ospfv2-ext:md5-authentication-key"]
    updated_md_key_list = []

    for index, message_digest in enumerate(intf["ip"]["ospf"]["message-digest-key"]):
        config = {
            "openconfig-ospfv2-ext:key-id": message_digest["id"],
            "openconfig-ospfv2-ext:config": {
                "openconfig-ospfv2-ext:key-id": message_digest["id"],
                "openconfig-ospfv2-ext:key": message_digest["md5"]["secret"]
            }
        }
        mess_digest_list.append(config)

        if "message-digest-key" in intf_leftover["ip"]["ospf"]:
            intf_leftover["ip"]["ospf"]["message-digest-key"][index] = None

    for md_key in intf_leftover.get("ip", {}).get("ospf", {}).get("message-digest-key", []):
        if md_key:
            updated_md_key_list.append(md_key)

    if len(updated_md_key_list) > 0:
        intf_leftover["ip"]["ospf"]["message-digest-key"] = updated_md_key_list
    elif "message-digest-key" in intf_leftover["ip"]["ospf"]:
        del intf_leftover["ip"]["ospf"]["message-digest-key"]

    return mess_digest


def set_ospfv2_areas(ospfv2_area, area, leftover_area, ospf, ospf_leftover):
    area_by_id = get_area_by_id(ospfv2_area, area["id"])
    set_traffic_eng(area_by_id, ospf, ospf_leftover)
    set_virtual_links(area_by_id, area, leftover_area)
    set_stub(area_by_id, area, leftover_area)


def set_traffic_eng(area_by_id, ospf, ospf_leftover):
    is_enabled = False
    if "mpls" in ospf and "traffic-eng" in ospf["mpls"] and "area" in ospf["mpls"]["traffic-eng"]:
        for traffic_area_index, traffic_area_id in enumerate(ospf["mpls"]["traffic-eng"]["area"]):
            if area_by_id["openconfig-network-instance:identifier"] == traffic_area_id:
                is_enabled = True
                ospf_leftover["mpls"]["traffic-eng"]["area"][traffic_area_index] = None
                break

    area_by_id["openconfig-network-instance:mpls"] = {
        "openconfig-network-instance:config": {"openconfig-network-instance:traffic-engineering-enabled": is_enabled}}


def set_virtual_links(area_by_id, area, leftover_area):
    if "virtual-link" in area:
        for v_link_index, v_link in enumerate(area["virtual-link"]):
            if not "openconfig-network-instance:virtual-links" in area_by_id:
                area_by_id["openconfig-network-instance:virtual-links"] = {
                    "openconfig-network-instance:virtual-link": []}
            if not "openconfig-network-instance:virtual-link" in area_by_id[
                "openconfig-network-instance:virtual-links"]:
                area_by_id["openconfig-network-instance:virtual-links"]["openconfig-network-instance:virtual-link"] = []

            area_by_id["openconfig-network-instance:virtual-links"]["openconfig-network-instance:virtual-link"].append({
                "openconfig-network-instance:remote-router-id": v_link["id"],
                "openconfig-network-instance:config": {"openconfig-network-instance:remote-router-id": v_link["id"]}
            })

            if len(leftover_area) > 0 and "virtual-link" in leftover_area[0]:
                leftover_area[0]["virtual-link"][v_link_index] = None


def set_stub(area_by_id, area, leftover_area):
    # This might not be necessary, since configs are coming directly from NEDs.
    # stub_counter = 0

    # if "stub" in area:
    #     stub_counter += 1

    #     if "no-summary" in area["stub"]:
    #         stub_counter += 1
    # if "nssa" in area:
    #     stub_counter += 1

    # if stub_counter > 1:
    #     raise ValueError("NED OSPF config has more than one stub type")
    all_true = {
        "openconfig-ospfv2-ext:enabled": True,
        "openconfig-ospfv2-ext:default-information-originate": True
    }
    all_false = {
        "openconfig-ospfv2-ext:enabled": False,
        "openconfig-ospfv2-ext:default-information-originate": False
    }

    if "stub" in area:
        nssa_false = copy.deepcopy(all_false)
        nssa_false["openconfig-ospfv2-ext:no-summary"] = False

        if "no-summary" in area["stub"]:
            area_by_id["openconfig-ospfv2-ext:stub-options"] = {
                "openconfig-ospfv2-ext:totally-stubby": {"openconfig-ospfv2-ext:config": copy.deepcopy(all_true)},
                "openconfig-ospfv2-ext:stub": {"openconfig-ospfv2-ext:config": copy.deepcopy(all_false)},
                "openconfig-ospfv2-ext:nssa": {"openconfig-ospfv2-ext:config": nssa_false}
            }
        else:
            area_by_id["openconfig-ospfv2-ext:stub-options"] = {
                "openconfig-ospfv2-ext:totally-stubby": {"openconfig-ospfv2-ext:config": copy.deepcopy(all_false)},
                "openconfig-ospfv2-ext:stub": {"openconfig-ospfv2-ext:config": copy.deepcopy(all_true)},
                "openconfig-ospfv2-ext:nssa": {"openconfig-ospfv2-ext:config": nssa_false}
            }

        if len(leftover_area) > 0 and "stub" in leftover_area[0]:
            del leftover_area[0]["stub"]
    elif "nssa" in area:
        nssa_true = copy.deepcopy(all_true)
        nssa_true["openconfig-ospfv2-ext:no-summary"] = "no-summary" in area["nssa"]
        area_by_id["openconfig-ospfv2-ext:stub-options"] = {
            "openconfig-ospfv2-ext:totally-stubby": {"openconfig-ospfv2-ext:config": copy.deepcopy(all_false)},
            "openconfig-ospfv2-ext:stub": {"openconfig-ospfv2-ext:config": copy.deepcopy(all_false)},
            "openconfig-ospfv2-ext:nssa": {"openconfig-ospfv2-ext:config": nssa_true}
        }

        if len(leftover_area) > 0 and "nssa" in leftover_area[0]:
            del leftover_area[0]["nssa"]



#! /usr/bin/env python3
"""
Translate NSO Device config to MDD OpenConfig

This script will pull a device's configuration from an NSO server, convert the NED structured configuration to
MDD OpenConfig, save the NSO configuration to a file named {device_name}_ned_configuration_interfaces.json, save the
NSO device configuration minus parts replaced by MDD OpenConfig to a file named
{device_name}_ned_configuration_remaining_interfaces.json, and save the MDD OpenConfig configuration to a file named
{nso_device}_openconfig_interfaces.json.

The script requires the following environment variables:
NSO_URL - URL for the NSO server
NSO_USERNAME
NSO_PASSWORD
NSO_DEVICE - NSO device name for configuration translation
TEST - True or False. True enables sending the OpenConfig to the NSO server after generation
"""
import sys
from pathlib import Path
from importlib.util import find_spec
import ipaddress

interfaces_notes = []

openconfig_interfaces = {
    "openconfig-interfaces:interfaces": {
        "openconfig-interfaces:interface": [
        ]
    }
}

nso_to_oc_interface_types = {
    "Ethernet": "ethernetCsmacd",
    "FastEthernet": "ethernetCsmacd",
    "FortyGigabitEthernet": "ethernetCsmacd",
    "GigabitEthernet": "ethernetCsmacd",
    "HundredGigE": "ethernetCsmacd",
    "TenGigabitEthernet": "ethernetCsmacd",
    "TwentyFiveGigE": "ethernetCsmacd",
    "TwoGigabitEthernet": "ethernetCsmacd",
    "Loopback": "softwareLoopback",
    "Tunnel": "tunnel",
    "vasileft": "vasi",
    "vasiright": "vasi",
    "Vlan": "l3ipvlan",
    "Port-channel": "ieee8023adLag",
    "Port-channel-subinterface": "ieee8023adLag"
}

port_speeds = {
    "10": "SPEED_10MB",
    "100": "SPEED_100MB",
    "1000": "SPEED_1GB",
    "2500": "SPEED_2500MB",
    "5000": "SPEED_5GB",
    "10000": "SPEED_10GB",
    "25000": "SPEED_25GB",
    "40000": "SPEED_40GB",
    "50000": "SPEED_50GB",
    "100000": "SPEED_100GB",
    "200000": "SPEED_200GB",
    "400000": "SPEED_400GB",
    "600000": "SPEED_600GB",
    "800000": "SPEED_800GB",
}

def interfaces_notes_add(note):
    interfaces_notes.append(note)


def return_nested_dict(root_dict: dict, keys_list: list) -> dict:
    """
    Return object of nested dictionary
    :param root_dict: full dict
    :param keys_list: list of keys; path to nested dict
    :return: dictionary object referencing nested dictionary
    """
    return_dict = root_dict
    for key in keys_list:
        return_dict = return_dict[key]
    return return_dict


def create_interface_dict(config_before: dict) -> dict:
    """
    Receive NSO configuration, create interface and index dict, and update global openconfig_interfaces
    Note - Interface type can be changed during processing
    :param config_before: dict
    :return: dictionaries of interface types, names, and indexes for nso and OC interface lists

    Example: {'Loopback': {'10': {'oc_interface_index': 0, 'nso_interface_index': 0, 'physical_interface_number': '10',
    'oc_sub_interface_number': 0, 'nso_interface_type': 'Loopback', 'oc_sub_interface_place_counter': 0},
    '100': {'oc_interface_index': 1, 'nso_interface_index': 1, 'physical_interface_number': '100',
    'oc_sub_interface_number': 0, 'nso_interface_type': 'Loopback', 'oc_sub_interface_place_counter': 0}},
    'GigabitEthernet': {'0/0': {'oc_interface_index': 2, 'nso_interface_index': 0, 'physical_interface_number': '0/0',
    'oc_sub_interface_number': 0, 'nso_interface_type': 'GigabitEthernet', 'oc_sub_interface_place_counter': 0},
    '0/1': {'oc_interface_index': 3, 'nso_interface_index': 1, 'physical_interface_number': '0/1',
    'oc_sub_interface_number': 0, 'nso_interface_type': 'GigabitEthernet', 'oc_sub_interface_place_counter': 0},
    '0/2': {'oc_interface_index': 4, 'nso_interface_index': 2, 'physical_interface_number': '0/2',
    'oc_sub_interface_number': 0, 'nso_interface_type': 'GigabitEthernet', 'oc_sub_interface_place_counter': 0},
    'Vlan': {100: {'oc_interface_index': 10, 'nso_interface_index': 0, 'physical_interface_number': '100',
    'oc_sub_interface_number': 0, 'nso_interface_type': 'Vlan', 'oc_sub_interface_place_counter': 0}}}
    """
    oc_interface_index = 0
    interface_dict = {}
    for interface_type in config_before.get("tailf-ned-cisco-ios:interface", {}):
        if interface_type != "Port-channel-subinterface" and nso_to_oc_interface_types.get(interface_type):
            interface_dict[interface_type] = {}
            nso_old_physical_interface_number = None
            oc_sub_interface_place_counter = 0  # OC interface sub-if place counter
            old_nso_index = 0  # Needed to not increase oc_interface_index when using subinterfaces
            for nso_index, value in enumerate(config_before["tailf-ned-cisco-ios:interface"][interface_type]):
                # Find sub-interface number
                interface_numbering = str(value["name"]).split('.')  # If '.' then number.sub-if
                physical_interface_number = interface_numbering[0]
                # sub-if is 0 unless there is a sub-if number
                oc_sub_interface_number = 0  # Most interfaces don't have a sub-if number
                if len(interface_numbering) > 1:  # If interface has a sub-if number then use that
                    oc_sub_interface_number = int(interface_numbering[1])

                # Are there more than one sub-if?
                # Note - the interface_dict is in order of physical and sub-interfaces
                # because they are organized that way in the NSO config, and we are iterating over NSO's config
                if physical_interface_number == nso_old_physical_interface_number:  # Means there is at least one sub-if
                    oc_sub_interface_place_counter += 1
                else:
                    oc_sub_interface_place_counter = 0

                if (nso_old_physical_interface_number == physical_interface_number) and (
                        nso_index == old_nso_index + 1):
                    oc_interface_index -= 1  # The OC interface index is for physical interfaces, if sub-if remove the past increase

                temp = {value["name"]:
                            {"oc_interface_index": oc_interface_index,
                             "nso_interface_index": nso_index,
                             "physical_interface_number": physical_interface_number,
                             "oc_sub_interface_number": oc_sub_interface_number,
                             "nso_interface_type": interface_type,
                             "oc_sub_interface_place_counter": oc_sub_interface_place_counter}
                        }
                interface_dict[interface_type].update(temp)

                if oc_sub_interface_place_counter == 0:
                    openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"].append(
                        {"openconfig-interfaces:name": f"{interface_type}{physical_interface_number}",
                         "openconfig-interfaces:config": {
                             "openconfig-interfaces:name": f"{interface_type}{physical_interface_number}"}})
                    oc_interface_index += 1  # Do not increase oc_interface_index for sub-interfaces
                old_nso_index = nso_index
                nso_old_physical_interface_number = physical_interface_number
                # When finished processing sub-ifs, increase oc_interface_index for next main/physical interface
                if oc_sub_interface_number > 0:
                    oc_interface_index += 1

        if interface_type == "Port-channel-subinterface" and config_before["tailf-ned-cisco-ios:interface"][
            interface_type].get("Port-channel") and nso_to_oc_interface_types.get(interface_type):
            interface_dict[interface_type] = {}
            oc_sub_interface_place_counter = 0  # OC interface sub-if place counter
            nso_old_physical_interface_number = None
            for nso_index, value in enumerate(
                    config_before["tailf-ned-cisco-ios:interface"]["Port-channel-subinterface"]["Port-channel"]):
                interface_numbering = str(value["name"]).split('.')  # If '.' then number.su-if

                physical_interface_number = interface_numbering[0]
                oc_sub_interface_number = int(interface_numbering[1])
                oc_interface_index = interface_dict["Port-channel"][int(physical_interface_number)][
                    "oc_interface_index"]

                if physical_interface_number == nso_old_physical_interface_number:  # Means there is at least one sub-if
                    oc_sub_interface_place_counter += 1
                else:
                    oc_sub_interface_place_counter = 0
                temp = {value["name"]:
                            {"oc_interface_index": oc_interface_index,
                             "nso_interface_index": nso_index,
                             "physical_interface_number": physical_interface_number,
                             "oc_sub_interface_number": oc_sub_interface_number,
                             "nso_interface_type": "Port-channel-subinterface",
                             "oc_sub_interface_place_counter": oc_sub_interface_place_counter}
                        }
                interface_dict[interface_type].update(temp)
                nso_old_physical_interface_number = physical_interface_number
            oc_interface_index += 1
    return interface_dict


def configure_switched_vlan(nso_before_interface: dict, nso_leftover_interface: dict,
                            openconfig_interface: dict, interface_name: str) -> None:
    """Configure L2 interfaces: TRUNK and ACCESS"""
    openconfig_interface.update(
        {"openconfig-vlan:switched-vlan": {"openconfig-vlan:config": {}}})
    # Mode ACCESS
    if type(nso_before_interface["switchport"].get("mode", {}).get("access")) is dict:
        openconfig_interface["openconfig-vlan:switched-vlan"][
            "openconfig-vlan:config"]["openconfig-vlan:interface-mode"] = "ACCESS"
        del nso_leftover_interface["switchport"]["mode"]
        openconfig_interface["openconfig-vlan:switched-vlan"][
            "openconfig-vlan:config"]["openconfig-vlan:access-vlan"] = \
            nso_before_interface["switchport"].get("access", {}).get("vlan")
        del nso_leftover_interface["switchport"]["access"]
    # Mode TRUNK
    elif (type(nso_before_interface["switchport"].get("mode", {}).get("trunk", "")) is dict):
        openconfig_interface["openconfig-vlan:switched-vlan"][
            "openconfig-vlan:config"]["openconfig-vlan:interface-mode"] = "TRUNK"
        del nso_leftover_interface["switchport"]["mode"]
        if nso_before_interface["switchport"].get("trunk", {}).get("native", {}).get("vlan"):
            openconfig_interface["openconfig-vlan:switched-vlan"][
                "openconfig-vlan:config"]["openconfig-vlan:native-vlan"] = \
                nso_before_interface["switchport"]["trunk"].get("native", {}).get("vlan")
            del nso_leftover_interface["switchport"]["trunk"]["native"]
        if nso_before_interface["switchport"].get("trunk", {}).get("allowed", {}).get("vlan", {}).get("vlans"):
            openconfig_interface["openconfig-vlan:switched-vlan"][
                "openconfig-vlan:config"]["openconfig-vlan:trunk-vlans"] = \
                nso_before_interface["switchport"].get("trunk", {}).get("allowed", {}).get("vlan", {}).get("vlans")
            del nso_leftover_interface["switchport"]["trunk"]["allowed"]
    # Mode dynamic: desirable or dynamic: auto will be a converted to TRUNK in OC
    elif nso_before_interface["switchport"].get("mode", {}).get("dynamic"):
        openconfig_interface["openconfig-vlan:switched-vlan"][
            "openconfig-vlan:config"]["openconfig-vlan:interface-mode"] = "TRUNK"
        del nso_leftover_interface["switchport"]["mode"]
        if nso_before_interface["switchport"].get("trunk", {}).get("native", {}).get("vlan"):
            openconfig_interface["openconfig-vlan:switched-vlan"][
                "openconfig-vlan:config"]["openconfig-vlan:native-vlan"] = \
                nso_before_interface["switchport"]["trunk"].get("native", {}).get("vlan")
            del nso_leftover_interface["switchport"]["trunk"]["native"]
        if nso_before_interface["switchport"].get("trunk", {}).get("allowed", {}).get("vlan", {}).get("vlans"):
            openconfig_interface["openconfig-vlan:switched-vlan"][
                "openconfig-vlan:config"]["openconfig-vlan:trunk-vlans"] = \
                nso_before_interface["switchport"].get("trunk", {}).get("allowed", {}).get("vlan", {}).get("vlans")
            del nso_leftover_interface["switchport"]["trunk"]["allowed"]
        interfaces_notes_add(f"""
            Interface {interface_name} was set to trunking dynamic {nso_before_interface["switchport"].get("mode", {}).get("dynamic")}.
            OpenConfig configuration is now mode TRUNK. Review for issues.
        """)


def xe_configure_ipv4_interface(nso_before_interface: dict, nso_leftover_interface: dict,
                                openconfig_interface: dict) -> None:
    """IPv4 interface configurations"""
    oc_ipv4_structure = {"openconfig-if-ip:ipv4": {"openconfig-if-ip:addresses": {"openconfig-if-ip:address": []},
                                                   "openconfig-if-ip:config": {}}}
    if (nso_before_interface.get("ip") and not nso_before_interface.get("ip", {}).get("no-address")) or (
            nso_before_interface.get("vrrp")):
        openconfig_interface.update(oc_ipv4_structure)
        ipv4_address_structure = {}
        if (nso_before_interface["ip"].get(
                "address", {}).get("primary", {}).get("address") and nso_before_interface["ip"].get("address", {}).get(
            "primary", {}).get("mask")):
            prefix = ipaddress.IPv4Network(
                f'{nso_before_interface["ip"].get("address", {}).get("primary", {}).get("address")}/{nso_before_interface["ip"].get("address", {}).get("primary", {}).get("mask")}',
                strict=False)
            mask = prefix.prefixlen
            ip = nso_before_interface["ip"].get("address", {}).get("primary", {}).get("address")
            del nso_leftover_interface["ip"]["address"]["primary"]
            ipv4_address_structure.update({"openconfig-if-ip:ip": ip,
                                           "openconfig-if-ip:config": {"openconfig-if-ip:ip": ip,
                                                                       "openconfig-if-ip:prefix-length": mask}})
        if len(ipv4_address_structure) > 0:
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:addresses"][
                "openconfig-if-ip:address"].append(ipv4_address_structure)
        if type(nso_before_interface["ip"].get(
                "address", {}).get("dhcp", "")) is dict:
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip:dhcp-client"] = True
            del \
                nso_before_interface["ip"]["address"]
        else:
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip:dhcp-client"] = False
        # VRRP
        if nso_before_interface.get("vrrp"):
            vrrp_dict = xe_configure_vrrp_interfaces(nso_before_interface, nso_leftover_interface)
            ipv4_address_structure.update(vrrp_dict)
        # HSRP
        if nso_before_interface.get("standby", {}).get("standby-list"):
            hsrp_dict = xe_configure_hsrp_interfaces(nso_before_interface, nso_leftover_interface)
            ipv4_address_structure.update(hsrp_dict)
        # IP MTU
        if nso_before_interface.get("ip", {}).get("mtu"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip:mtu"] = nso_before_interface.get("ip", {}).get("mtu")

            del nso_leftover_interface["ip"]["mtu"]
        # adjust TCP MSS
        if nso_before_interface.get("ip", {}).get("tcp", {}).get("adjust-mss"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip-mdd-ext:tcp-adjust-mss"] = nso_before_interface["ip"]["tcp"]["adjust-mss"]

            del nso_leftover_interface["ip"]["tcp"]["adjust-mss"]
        # IP redirects
        if nso_before_interface.get("ip", {}).get("redirects"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip-mdd-ext:redirects"] = True
            del nso_leftover_interface["ip"]["redirects"]
        elif nso_before_interface.get("ip", {}).get("redirects") is False:
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip-mdd-ext:redirects"] = False
            del nso_leftover_interface["ip"]["redirects"]
        # IP unreachables
        if nso_before_interface.get("ip", {}).get("unreachables"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip-mdd-ext:unreachables"] = True
            del nso_leftover_interface["ip"]["unreachables"]
        elif nso_before_interface.get("ip", {}).get(
                "unreachables") is False:
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip-mdd-ext:unreachables"] = False
            del nso_leftover_interface["ip"]["unreachables"]
        # Proxy-ARP
        if nso_before_interface.get("ip", {}).get("proxy-arp"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:proxy-arp"] = {
                "openconfig-if-ip:config": {"openconfig-if-ip:mode": "REMOTE_ONLY"}}
            del nso_leftover_interface["ip"]["proxy-arp"]
        elif nso_before_interface.get("ip", {}).get("proxy-arp") is False:
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:proxy-arp"] = {
                "openconfig-if-ip:config": {"openconfig-if-ip:mode": "DISABLE"}}
            del nso_leftover_interface["ip"]["proxy-arp"]
        # reply-mask
        if nso_before_interface.get("ip", {}).get("mask-reply"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"][
                "openconfig-if-ip-mdd-ext:mask-reply"] = True
            del nso_leftover_interface["ip"]["mask-reply"]
        # NAT interface
        if nso_before_interface.get("ip", {}).get("nat", {}).get("inside"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"]["openconfig-if-ip-mdd-ext:nat"] = {
                "openconfig-if-ip-mdd-ext:nat-choice": "inside"}
            del nso_leftover_interface["ip"]["nat"]["inside"]
        elif nso_before_interface.get("ip", {}).get("nat", {}).get("outside"):
            openconfig_interface["openconfig-if-ip:ipv4"]["openconfig-if-ip:config"]["openconfig-if-ip-mdd-ext:nat"] = {
                "openconfig-if-ip-mdd-ext:nat-choice": "outside"}
            del nso_leftover_interface["ip"]["nat"]["outside"]


def xe_configure_tunnel_ipv4_interface(nso_before_interface: dict, nso_leftover_interface: dict,
                                       openconfig_interface: dict) -> None:
    """Tunnel IPv4 interface configurations"""
    oc_ipv4_structure = {
        "openconfig-if-tunnel:ipv4": {"openconfig-if-tunnel:addresses": {"openconfig-if-tunnel:address": []},
                                      "openconfig-if-tunnel:config": {}}}
    if (nso_before_interface.get("ip") and not nso_before_interface.get("ip", {}).get("no-address")) or (
            nso_before_interface.get("vrrp")):
        openconfig_interface.update(oc_ipv4_structure)
        ipv4_address_structure = {}
        if (nso_before_interface["ip"].get(
                "address", {}).get("primary", {}).get("address") and nso_before_interface["ip"].get("address", {}).get(
            "primary", {}).get("mask")):
            prefix = ipaddress.IPv4Network(
                f'{nso_before_interface["ip"].get("address", {}).get("primary", {}).get("address")}/{nso_before_interface["ip"].get("address", {}).get("primary", {}).get("mask")}',
                strict=False)
            mask = prefix.prefixlen
            ip = nso_before_interface["ip"].get("address", {}).get("primary", {}).get("address")
            del nso_leftover_interface["ip"]["address"]["primary"]
            ipv4_address_structure.update({"openconfig-if-tunnel:ip": ip,
                                           "openconfig-if-tunnel:config": {"openconfig-if-tunnel:ip": ip,
                                                                           "openconfig-if-tunnel:prefix-length": mask}})
        if len(ipv4_address_structure) > 0:
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:addresses"][
                "openconfig-if-tunnel:address"].append(ipv4_address_structure)
        if type(nso_before_interface["ip"].get(
                "address", {}).get("dhcp", "")) is dict:
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel:dhcp-client"] = True
            del \
                nso_before_interface["ip"]["address"]
        else:
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel:dhcp-client"] = False
        # VRRP
        if nso_before_interface.get("vrrp"):
            vrrp_dict = xe_configure_vrrp_interfaces(nso_before_interface, nso_leftover_interface)
            ipv4_address_structure.update(vrrp_dict)
        # HSRP
        if nso_before_interface.get("standby", {}).get("standby-list"):
            hsrp_dict = xe_configure_hsrp_interfaces(nso_before_interface, nso_leftover_interface)
            ipv4_address_structure.update(hsrp_dict)
        # IP MTU
        if nso_before_interface.get("ip", {}).get("mtu"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel:mtu"] = nso_before_interface.get("ip", {}).get("mtu")

            del nso_leftover_interface["ip"]["mtu"]
        # adjust TCP MSS
        if nso_before_interface.get("ip", {}).get("tcp", {}).get("adjust-mss"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:tcp-adjust-mss"] = nso_before_interface["ip"]["tcp"]["adjust-mss"]

            del nso_leftover_interface["ip"]["tcp"]["adjust-mss"]
        # IP redirects
        if nso_before_interface.get("ip", {}).get("redirects"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:redirects"] = True
            del nso_leftover_interface["ip"]["redirects"]
        elif nso_before_interface.get("ip", {}).get("redirects") is False:
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:redirects"] = False
            del nso_leftover_interface["ip"]["redirects"]
        # IP unreachables
        if nso_before_interface.get("ip", {}).get("unreachables"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:unreachables"] = True
            del nso_leftover_interface["ip"]["unreachables"]
        elif nso_before_interface.get("ip", {}).get(
                "unreachables") is False:
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:unreachables"] = False
            del nso_leftover_interface["ip"]["unreachables"]
        # Proxy-ARP
        if nso_before_interface.get("ip", {}).get("proxy-arp"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:proxy-arp"] = {
                "openconfig-if-tunnel:config": {"openconfig-if-tunnel:mode": "REMOTE_ONLY"}}
            del nso_leftover_interface["ip"]["proxy-arp"]
        elif nso_before_interface.get("ip", {}).get("proxy-arp") is False:
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:proxy-arp"] = {
                "openconfig-if-tunnel:config": {"openconfig-if-tunnel:mode": "DISABLE"}}
            del nso_leftover_interface["ip"]["proxy-arp"]
        # reply-mask
        if nso_before_interface.get("ip", {}).get("mask-reply"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:mask-reply"] = True
            del nso_leftover_interface["ip"]["mask-reply"]
        # NAT interface
        if nso_before_interface.get("ip", {}).get("nat", {}).get("inside"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:nat"] = {
                "openconfig-if-ip-mdd-ext:nat-choice": "inside"}
            del nso_leftover_interface["ip"]["nat"]["inside"]
        elif nso_before_interface.get("ip", {}).get("nat", {}).get("outside"):
            openconfig_interface["openconfig-if-tunnel:ipv4"]["openconfig-if-tunnel:config"][
                "openconfig-if-ip-mdd-ext:nat"] = {
                "openconfig-if-ip-mdd-ext:nat-choice": "outside"}
            del nso_leftover_interface["ip"]["nat"]["outside"]


def configure_software_loopback(config_before: dict, config_leftover: dict, interface_data: dict) -> None:
    """Configure Loopbacks"""
    for interface_directory in interface_data.values():
        path_oc_sub_if = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                          interface_directory["oc_interface_index"], "openconfig-interfaces:subinterfaces",
                          "openconfig-interfaces:subinterface", interface_directory["oc_sub_interface_place_counter"]]
        path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                   interface_directory["oc_interface_index"]]
        path_nso = ["tailf-ned-cisco-ios:interface", interface_directory["nso_interface_type"],
                    interface_directory["nso_interface_index"]]
        openconfig_interface_sub_if = return_nested_dict(openconfig_interfaces, path_oc_sub_if)
        openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
        nso_before_interface = return_nested_dict(config_before, path_nso)
        nso_leftover_interface = return_nested_dict(config_leftover, path_nso)

        # Main Interface
        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface)
        # Sub Interface
        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface_sub_if)
        xe_configure_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface_sub_if)


def configure_software_vasi(config_before: dict, config_leftover: dict, interface_data: dict) -> None:
    """Configure VASI"""
    for interface_directory in interface_data.values():
        path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                   interface_directory["oc_interface_index"], "openconfig-interfaces:subinterfaces",
                   "openconfig-interfaces:subinterface", interface_directory["oc_sub_interface_place_counter"]]
        path_nso = ["tailf-ned-cisco-ios:interface", interface_directory["nso_interface_type"],
                    interface_directory["nso_interface_index"]]
        openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
        nso_before_interface = return_nested_dict(config_before, path_nso)
        nso_leftover_interface = return_nested_dict(config_leftover, path_nso)

        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface)
        xe_configure_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface)


def configure_software_l3ipvlan(config_before: dict, config_leftover: dict, interface_data: dict) -> None:
    """Configure routed VLANs"""
    for interface_directory in interface_data.values():
        path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                   interface_directory["oc_interface_index"]]
        path_nso = ["tailf-ned-cisco-ios:interface", interface_directory["nso_interface_type"],
                    interface_directory["nso_interface_index"]]
        openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
        nso_before_interface = return_nested_dict(config_before, path_nso)
        nso_leftover_interface = return_nested_dict(config_leftover, path_nso)

        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface)
        vlan = nso_before_interface.get("name")

        openconfig_interface.update(
            {"openconfig-vlan:routed-vlan": {"openconfig-vlan:config": {"openconfig-vlan:vlan": vlan}}})

        path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                   interface_directory["oc_interface_index"], "openconfig-vlan:routed-vlan"]
        openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
        xe_configure_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface)


def configure_port_channel(config_before: dict, config_leftover: dict, interface_data: dict) -> None:
    """Configure LACP port-channel"""
    for interface_directory in interface_data.values():
        # Configure port-channel interface
        if interface_directory["nso_interface_type"] == "Port-channel":
            path_oc_physical = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                                interface_directory["oc_interface_index"]]
            openconfig_interface_physical = return_nested_dict(openconfig_interfaces, path_oc_physical)
            interface_name = interface_directory["nso_interface_type"] + str(
                config_before["tailf-ned-cisco-ios:interface"][interface_directory["nso_interface_type"]][
                    interface_directory["nso_interface_index"]]["name"])
            path_nso_physical = ["tailf-ned-cisco-ios:interface", interface_directory["nso_interface_type"],
                                 interface_directory["nso_interface_index"]]
            nso_before_interface = return_nested_dict(config_before, path_nso_physical)
            nso_leftover_interface = return_nested_dict(config_leftover, path_nso_physical)
            xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface_physical)
            mtu_set(nso_before_interface, nso_leftover_interface, openconfig_interface_physical)
            openconfig_interface_physical.update({"openconfig-if-aggregate:aggregation": {
                "openconfig-if-aggregate:config": {"openconfig-if-aggregate:lag-type": "LACP"}}})
            path_oc_agg = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                           interface_directory["oc_interface_index"], "openconfig-if-aggregate:aggregation"]
            openconfig_interface_agg = return_nested_dict(openconfig_interfaces, path_oc_agg)
            if nso_before_interface.get("switchport"):
                configure_switched_vlan(nso_before_interface, nso_leftover_interface, openconfig_interface_agg,
                                        interface_name)
            xe_configure_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface_agg)
        # Configure port-channel sub-interfaces
        if interface_directory["nso_interface_type"] == "Port-channel-subinterface":
            path_nso_subif = ["tailf-ned-cisco-ios:interface", "Port-channel-subinterface", "Port-channel",
                              interface_directory["nso_interface_index"]]
            nso_before_interface = return_nested_dict(config_before, path_nso_subif)
            nso_leftover_interface = return_nested_dict(config_leftover, path_nso_subif)
            path_oc_physical = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                                interface_directory["oc_interface_index"]]
            openconfig_interface_physical = return_nested_dict(openconfig_interfaces, path_oc_physical)
            if not openconfig_interface_physical.get("openconfig-interfaces:subinterfaces"):
                openconfig_interface_physical.update(
                    {"openconfig-interfaces:subinterfaces": {"openconfig-interfaces:subinterface": []}})

            openconfig_interface_physical["openconfig-interfaces:subinterfaces"][
                "openconfig-interfaces:subinterface"].append(
                {"openconfig-interfaces:index": interface_directory["oc_sub_interface_number"],
                 "openconfig-interfaces:config": {
                     "openconfig-interfaces:index": interface_directory[
                         "oc_sub_interface_number"]}})

            path_oc_subif = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                             interface_directory["oc_interface_index"], "openconfig-interfaces:subinterfaces",
                             "openconfig-interfaces:subinterface",
                             interface_directory["oc_sub_interface_place_counter"]]

            openconfig_interface_subif = return_nested_dict(openconfig_interfaces, path_oc_subif)

            xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface_subif)
            xe_configure_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface_subif)


def check_for_ip_address(address):
    try:
        ipaddress.ip_address(address)
        return True
    except ValueError:
        return False


def configure_software_tunnel(config_before: dict, config_leftover: dict, interface_data: dict, if_ip: dict) -> None:
    """Configure GRE Tunnel"""
    for interface_directory in interface_data.values():
        # Configure tunnel interface
        path_oc_physical = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                            interface_directory["oc_interface_index"]]
        openconfig_interface_physical = return_nested_dict(openconfig_interfaces, path_oc_physical)
        path_nso_physical = ["tailf-ned-cisco-ios:interface", interface_directory["nso_interface_type"],
                             interface_directory["nso_interface_index"]]
        nso_before_interface = return_nested_dict(config_before, path_nso_physical)
        nso_leftover_interface = return_nested_dict(config_leftover, path_nso_physical)

        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface_physical)

        openconfig_interface_physical["openconfig-if-tunnel:tunnel"] = {"openconfig-if-tunnel:config": {}}
        path_oc_tunnel = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                          interface_directory["oc_interface_index"], "openconfig-if-tunnel:tunnel"]
        openconfig_interface_tunnel = return_nested_dict(openconfig_interfaces, path_oc_tunnel)

        xe_configure_tunnel_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface_tunnel)

        # source IP
        if nso_before_interface.get("tunnel", {}).get("source"):
            tunnel_src = nso_before_interface.get("tunnel", {}).get("source")
            if check_for_ip_address(tunnel_src):
                openconfig_interface_tunnel["openconfig-if-tunnel:config"][
                    "openconfig-if-tunnel:src"] = tunnel_src
            else:
                openconfig_interface_tunnel["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel:src"] = if_ip.get(tunnel_src)
            del nso_leftover_interface["tunnel"]["source"]
        # destination IP
        if nso_before_interface.get("tunnel", {}).get("destination"):
            openconfig_interface_tunnel["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel:dst"] = nso_before_interface.get("tunnel", {}).get("destination")
            del nso_leftover_interface["tunnel"]["destination"]
        # key
        if nso_before_interface.get("tunnel", {}).get("key"):
            openconfig_interface_tunnel["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel:gre-key"] = nso_before_interface.get("tunnel", {}).get("key")
            del nso_leftover_interface["tunnel"]["key"]
        # PMTUD
        if type(nso_before_interface.get("tunnel", {}).get("path-mtu-discovery", "")) is dict:
            openconfig_interface_tunnel["openconfig-if-tunnel:config"][
                "openconfig-if-tunnel-ext:tunnel-path-mtu-discovery"] = True
            del nso_leftover_interface["tunnel"]["path-mtu-discovery"]
        # keepalives
        if nso_before_interface.get("keepalive-period-retries", {}).get("keepalive", {}).get(
                "period") and nso_before_interface.get("keepalive-period-retries", {}).get("keepalive", {}).get(
            "retries"):
            openconfig_interface_tunnel["openconfig-if-tunnel:config"]["openconfig-if-tunnel-ext:keepalives"] = {
                "openconfig-if-tunnel-ext:period": nso_before_interface.get("keepalive-period-retries", {}).get(
                    "keepalive", {}).get("period"),
                "openconfig-if-tunnel-ext:retries": nso_before_interface.get("keepalive-period-retries", {}).get(
                    "keepalive", {}).get("retries")}
            del nso_leftover_interface["keepalive-period-retries"]


def mtu_set(nso_before_interface: dict, nso_leftover_interface: dict, openconfig_interface: dict) -> None:
    """
    Only on physical interfaces.
    Configures interface MTU.
    """
    # MTU
    if nso_before_interface.get("mtu"):
        openconfig_interface["openconfig-interfaces:config"]["openconfig-interfaces:mtu"] = nso_before_interface.get(
            "mtu")
        try:
            del nso_leftover_interface["mtu"]
        except:
            pass


def xe_interface_encapsulation(nso_before_interface: dict, nso_leftover_interface: dict, openconfig_interface: dict) -> None:
    # Encapsulation
    if nso_before_interface.get("encapsulation", {}).get("dot1Q", {}).get("vlan-id"):
        openconfig_interface.update(
            {"openconfig-vlan:vlan": {"openconfig-vlan:config": {
                "openconfig-vlan:vlan-id": nso_before_interface.get("encapsulation", {}).get("dot1Q", {}).get(
                    "vlan-id")}}})
        del nso_leftover_interface["encapsulation"]


def xe_interface_config(nso_before_interface: dict, nso_leftover_interface: dict, openconfig_interface: dict) -> None:
    """
    Configure basic interface functions, i.e. description, shutdown, MTU
    Note - subinterface 0 values are removed from config leftover during configuration of the physical interface
    """
    # Description
    if nso_before_interface.get('description'):
        openconfig_interface["openconfig-interfaces:config"][
            'openconfig-interfaces:description'] = nso_before_interface.get('description')
        try:
            del nso_leftover_interface["description"]
        except:
            pass
    # Shutdown
    if nso_before_interface.get('shutdown'):
        openconfig_interface["openconfig-interfaces:config"]["openconfig-interfaces:enabled"] = False
        try:
            del nso_leftover_interface["shutdown"]
        except:
            pass
    else:
        openconfig_interface["openconfig-interfaces:config"]["openconfig-interfaces:enabled"] = True


def xe_interface_hold_time(config_before: dict, config_leftover: dict, v: dict) -> None:
    """Configure physical interface hold-time (carrier-delay)"""
    if config_before["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]].get(
            "carrier-delay", {}).get("msec"):
        openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"][
            v["oc_interface_index"]]["openconfig-interfaces:hold-time"] = {"openconfig-interfaces:config": {
            "openconfig-interfaces:down": config_before["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][
                v["nso_interface_index"]].get("carrier-delay", {}).get("msec")}}
        del config_leftover["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]][
            "carrier-delay"]["msec"]


def xe_interface_storm_control(openconfig_interface: dict, nso_before_interface: dict, config_leftover: dict, v: dict) -> None:
    """Configure physical interface storm control"""

    openconfig_interface.update({"openconfig-if-ethernet:ethernet": {
        "openconfig-if-ethernet-mdd-ext:storm-control": {
            "openconfig-if-ethernet-mdd-ext:broadcast": {
                "openconfig-if-ethernet-mdd-ext:level": {
                    "openconfig-if-ethernet-mdd-ext:config": {}
                }
            },
            "openconfig-if-ethernet-mdd-ext:unicast": {
                "openconfig-if-ethernet-mdd-ext:level": {
                    "openconfig-if-ethernet-mdd-ext:config": {}
                }
            }
        }
    }})
    # broadcast
    if nso_before_interface.get("storm-control", {}).get("broadcast", {}).get("level-bps-pps", {}).get("level", {}).get("bps"):
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:broadcast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:suppression-type"] = 'BPS'
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:broadcast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:bps"] = nso_before_interface.get("storm-control", {}).get("broadcast", {}).get("level-bps-pps", {}).get("level", {}).get("bps")
        del config_leftover["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]][
                "storm-control"]["broadcast"]["level-bps-pps"]["level"]["bps"]
    elif nso_before_interface.get("storm-control", {}).get("broadcast", {}).get("level-bps-pps", {}).get("level", {}).get("pps"):
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:broadcast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:suppression-type"] = 'PPS'
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:broadcast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:pps"] = nso_before_interface.get("storm-control", {}).get("broadcast", {}).get("level-bps-pps", {}).get("level", {}).get("pps")
        del config_leftover["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]][
                "storm-control"]["broadcast"]["level-bps-pps"]["level"]["pps"]
    # unicast
    if nso_before_interface.get("storm-control", {}).get("unicast", {}).get("level-bps-pps", {}).get("level", {}).get("bps"):
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:unicast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:suppression-type"] = 'BPS'
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:unicast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:bps"] = nso_before_interface.get("storm-control", {}).get("unicast", {}).get("level-bps-pps", {}).get("level", {}).get("bps")
        del config_leftover["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]][
                "storm-control"]["unicast"]["level-bps-pps"]["level"]["bps"]
    elif nso_before_interface.get("storm-control", {}).get("unicast", {}).get("level-bps-pps", {}).get("level", {}).get("pps"):
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:unicast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:suppression-type"] = 'PPS'
        openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet-mdd-ext:storm-control"][
            "openconfig-if-ethernet-mdd-ext:unicast"]["openconfig-if-ethernet-mdd-ext:level"][
            "openconfig-if-ethernet-mdd-ext:config"]["openconfig-if-ethernet-mdd-ext:pps"] = nso_before_interface.get("storm-control", {}).get("unicast", {}).get("level-bps-pps", {}).get("level", {}).get("pps")
        del config_leftover["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]][
                "storm-control"]["unicast"]["level-bps-pps"]["level"]["pps"]


def xe_configure_vrrp_interfaces(nso_before_interface: dict, nso_leftover_interface: dict) -> dict:
    """Configure VRRP"""
    updated_vrrp = []

    service_vrrp = {"openconfig-if-ip:vrrp": {"openconfig-if-ip:vrrp-group": []}}
    for number, group in enumerate(nso_before_interface.get("vrrp")):
        if group.get("id"):
            # Group
            service_vrrp_group = {"openconfig-if-ip:virtual-router-id": group.get("id"),
                                  "openconfig-if-ip:config": {"openconfig-if-ip:virtual-router-id": group.get("id")}}
            del nso_leftover_interface["vrrp"][number]["id"]
            # Preempt delay
            if group.get("preempt", {}).get("delay", {}).get("minimum"):
                service_vrrp_group["openconfig-if-ip:config"]["openconfig-if-ip:preempt-delay"] = group.get("preempt",
                                                                                                            {}).get(
                    "delay", {}).get("minimum")
                del nso_leftover_interface["vrrp"][number]["preempt"]["delay"]
            # Preempt
            if group.get("preempt"):
                service_vrrp_group["openconfig-if-ip:config"]["openconfig-if-ip:preempt"] = True
                del nso_leftover_interface["vrrp"][number]["preempt"]
            # Priority
            if group.get("priority"):
                service_vrrp_group["openconfig-if-ip:config"]["openconfig-if-ip:priority"] = group.get("priority")
                del nso_leftover_interface["vrrp"][number]["priority"]
            # VRRP Address
            if group.get("ip", {}).get("address"):
                service_vrrp_group["openconfig-if-ip:config"]["openconfig-if-ip:virtual-address"] = []
                service_vrrp_group["openconfig-if-ip:config"]["openconfig-if-ip:virtual-address"].append(
                    group.get("ip", {}).get("address"))
                del nso_leftover_interface["vrrp"][number]["ip"]
            # Timers advertise
            if group.get("timers", {}).get("advertise", {}).get("seconds"):
                service_vrrp_group["openconfig-if-ip:config"]["openconfig-if-ip:advertisement-interval"] = int(
                    group.get("timers", {}).get("advertise").get("seconds")) * 100
                del nso_leftover_interface["vrrp"][number]["timers"]["advertise"]
            service_vrrp["openconfig-if-ip:vrrp"]["openconfig-if-ip:vrrp-group"].append(service_vrrp_group)
            # Clean up
            if nso_leftover_interface["vrrp"][number] and len(nso_leftover_interface["vrrp"][number]) > 0:
                updated_vrrp.append(nso_leftover_interface["vrrp"][number])
    nso_leftover_interface["vrrp"] = updated_vrrp
    
    return service_vrrp


def xe_configure_hsrp_interfaces(nso_before_interface: dict, nso_leftover_interface: dict) -> dict:
    """Configure HSRP"""
    service_hsrp = {"openconfig-if-ip-mdd-ext:hsrp": {"openconfig-if-ip-mdd-ext:hsrp-group": []}}
    for number, group in enumerate(nso_before_interface.get("standby", {}).get("standby-list")):
        if group.get("group-number"):
            # Group
            service_hsrp_group = {"openconfig-if-ip-mdd-ext:group-number": group.get("group-number"),
                                  "openconfig-if-ip-mdd-ext:config": {
                                      "openconfig-if-ip-mdd-ext:group-number": group.get("group-number")}}
            del nso_leftover_interface["standby"]["standby-list"][number]["group-number"]
            # Preempt delay
            if group.get("preempt", {}).get("delay", {}).get("minimum"):
                service_hsrp_group["openconfig-if-ip-mdd-ext:config"][
                    "openconfig-if-ip-mdd-ext:preempt-delay"] = group.get("preempt",
                                                                          {}).get(
                    "delay", {}).get("minimum")
                del nso_leftover_interface["standby"]["standby-list"][number]["preempt"]["delay"]
            # Preempt
            if group.get("preempt"):
                service_hsrp_group["openconfig-if-ip-mdd-ext:config"]["openconfig-if-ip-mdd-ext:preempt"] = True
                del nso_leftover_interface["standby"]["standby-list"][number]["preempt"]
            # Priority
            if group.get("priority"):
                service_hsrp_group["openconfig-if-ip-mdd-ext:config"]["openconfig-if-ip-mdd-ext:priority"] = group.get(
                    "priority")
                del nso_leftover_interface["standby"]["standby-list"][number]["priority"]
            # VRRP Address
            if group.get("ip", {}).get("address"):
                service_hsrp_group["openconfig-if-ip-mdd-ext:config"]["openconfig-if-ip-mdd-ext:virtual-address"] = []
                service_hsrp_group["openconfig-if-ip-mdd-ext:config"][
                    "openconfig-if-ip-mdd-ext:virtual-address"].append(
                    group.get("ip", {}).get("address"))
                del nso_leftover_interface["standby"]["standby-list"][number]["ip"]
            # Timers
            if group.get("timers", {}).get("hello-interval", {}).get("seconds") and group.get("timers", {}).get(
                    "hold-time", {}).get("seconds"):
                service_hsrp_group["openconfig-if-ip-mdd-ext:config"].update({"openconfig-if-ip-mdd-ext:timers": {
                    "openconfig-if-ip-mdd-ext:hello-interval": int(
                        group.get("timers", {}).get("hello-interval").get("seconds")),
                    "openconfig-if-ip-mdd-ext:holdtime": int(group.get("timers", {}).get("hold-time").get("seconds"))
                }})
                del nso_leftover_interface["standby"]["standby-list"][number]["timers"]["hello-interval"]
                del nso_leftover_interface["standby"]["standby-list"][number]["timers"]["hold-time"]

            service_hsrp["openconfig-if-ip-mdd-ext:hsrp"]["openconfig-if-ip-mdd-ext:hsrp-group"].append(
                service_hsrp_group)
    return service_hsrp


def configure_csmacd(config_before: dict, config_leftover: dict, interface_data: dict) -> None:
    """
    Iterate through interface_data
    Call up the config_before["tailf-ned-cisco-ios:interface"][v["nso_interface_type"]][v["nso_interface_index"]]
    Add need OC config to openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"][v["oc_interface_index"]]
    """
    for interface_directory in interface_data.values():
        path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                   interface_directory["oc_interface_index"], "openconfig-interfaces:subinterfaces",
                   "openconfig-interfaces:subinterface", interface_directory["oc_sub_interface_place_counter"]]
        openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
        interface_name = interface_directory["nso_interface_type"] + str(
            config_before["tailf-ned-cisco-ios:interface"][interface_directory["nso_interface_type"]][
                interface_directory["nso_interface_index"]]["name"])
        path_nso = ["tailf-ned-cisco-ios:interface", interface_directory["nso_interface_type"],
                    interface_directory["nso_interface_index"]]
        nso_before_interface = return_nested_dict(config_before, path_nso)
        nso_leftover_interface = return_nested_dict(config_leftover, path_nso)

        # Configure sub-interface
        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface)
        xe_interface_encapsulation(nso_before_interface, nso_leftover_interface, openconfig_interface)

        path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                   interface_directory["oc_interface_index"]]
        openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)

        # Configure physical interface
        xe_interface_config(nso_before_interface, nso_leftover_interface, openconfig_interface)
        mtu_set(nso_before_interface, nso_leftover_interface, openconfig_interface)
        # Configure physical interface hold-time (carrier-delay)
        xe_interface_hold_time(config_before, config_leftover, interface_directory)

        # Configure ethernet settings
        openconfig_interface.update({"openconfig-if-ethernet:ethernet": {"openconfig-if-ethernet:config": {}}})
        if nso_before_interface.get("storm-control"):
            xe_interface_storm_control(openconfig_interface, nso_before_interface, config_leftover, interface_directory)
        if nso_before_interface.get("negotiation", {}).get("auto"):
            openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet:config"][
                "openconfig-if-ethernet:auto-negotiate"] = True
            del nso_leftover_interface["negotiation"]["auto"]
        elif (nso_before_interface.get("speed") == "auto") and (nso_before_interface.get("duplex") == "auto"):
            openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet:config"][
            "openconfig-if-ethernet:auto-negotiate"] = True
            del nso_leftover_interface["speed"]
            del nso_leftover_interface["duplex"]
        else:
            if nso_before_interface.get("speed", "").isdigit():
                if port_speeds.get(nso_before_interface.get("speed")):
                    openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet:config"][
                        "openconfig-if-ethernet:port-speed"] = port_speeds.get(nso_before_interface.get("speed"))
                    del nso_leftover_interface["speed"]
            if nso_before_interface.get("duplex") == "full" or nso_before_interface.get("duplex") == "half":
                openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet:config"][
                    "openconfig-if-ethernet:duplex-mode"] = nso_before_interface.get("duplex").upper()
                del nso_leftover_interface["duplex"]
        if nso_before_interface.get("flowcontrol", {}).get("receive") == "on":
            openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet:config"][
                "openconfig-if-ethernet:enable-flow-control"] = True
            del nso_leftover_interface["flowcontrol"]["receive"]
        if nso_before_interface.get("mac-address"):
            n_mac = nso_before_interface.get("mac-address")
            openconfig_interface["openconfig-if-ethernet:ethernet"]["openconfig-if-ethernet:config"][
                "openconfig-interfaces:mac-address"] = f"{n_mac[0:2]}:{n_mac[2:4]}:{n_mac[5:7]}:{n_mac[7:9]}:{n_mac[10:12]}:{n_mac[12:14]}"  # NSO 5254.0014.3427 OC = MM:MM:MM:SS:SS:SS
            del nso_leftover_interface["mac-address"]

        # Is type really a l2vlan?
        if nso_before_interface.get("switchport"):
            openconfig_interface["openconfig-interfaces:config"]["openconfig-interfaces:type"] = "l2vlan"
            path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                       interface_directory["oc_interface_index"], "openconfig-if-ethernet:ethernet"]
            openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
            configure_switched_vlan(nso_before_interface, nso_leftover_interface, openconfig_interface, interface_name)
        else:
            path_oc = ["openconfig-interfaces:interfaces", "openconfig-interfaces:interface",
                       interface_directory["oc_interface_index"], "openconfig-interfaces:subinterfaces",
                       "openconfig-interfaces:subinterface", interface_directory["oc_sub_interface_place_counter"]]
            openconfig_interface = return_nested_dict(openconfig_interfaces, path_oc)
            xe_configure_ipv4_interface(nso_before_interface, nso_leftover_interface, openconfig_interface)

        # Is interface an LACP member?
        if (nso_before_interface.get("channel-group", {}).get("number")) and (
                nso_before_interface.get("channel-group", {}).get("mode", "") == "active"):
            openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"][
                interface_directory["oc_interface_index"]]["openconfig-if-ethernet:ethernet"][
                "openconfig-if-ethernet:config"][
                "openconfig-if-aggregate:aggregate-id"] = f'Port-channel{str(nso_before_interface.get("channel-group", {}).get("number"))}'
            del nso_leftover_interface["channel-group"]


def xe_interfaces(config_before: dict, config_leftover: dict, interfaces: dict, if_ip: dict) -> None:
    """
    Translates NSO XE NED to MDD OpenConfig Interfaces Config
    """
    # Identify interface types
    for interface_type in interfaces:
        for interface in interfaces[interface_type]:
            # Assign default interface types
            interfaces[interface_type][interface]["oc_type"] = nso_to_oc_interface_types.get(interface_type,
                                                                                             "unsupported")
            openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"][
                interfaces[interface_type][interface]["oc_interface_index"]]["openconfig-interfaces:config"][
                "openconfig-interfaces:type"] = nso_to_oc_interface_types[interface_type]

    for interface_type in interfaces:
        for v in interfaces[interface_type].values():
            if_types_with_subifs = ["ethernetCsmacd", "softwareLoopback", "vasi"]
            if v["oc_type"] in if_types_with_subifs:
                if v["oc_sub_interface_number"] == 0:
                    subif = {"openconfig-interfaces:subinterfaces": {"openconfig-interfaces:subinterface": []}}
                    openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"][
                        v["oc_interface_index"]].update(subif)
                temp = {"openconfig-interfaces:index": v["oc_sub_interface_number"],
                        "openconfig-interfaces:config": {"openconfig-interfaces:index": v["oc_sub_interface_number"]}}
                openconfig_interfaces["openconfig-interfaces:interfaces"]["openconfig-interfaces:interface"][
                    v["oc_interface_index"]]["openconfig-interfaces:subinterfaces"][
                    "openconfig-interfaces:subinterface"].append(temp)

    # Configure the interface types
    for interface_type in interfaces:
        if nso_to_oc_interface_types[interface_type] == "ieee8023adLag":
            configure_port_channel(config_before, config_leftover, interfaces[interface_type])
        if nso_to_oc_interface_types[interface_type] == "ethernetCsmacd":
            # can be converted to l2vlan if contains 'switchport'
            configure_csmacd(config_before, config_leftover, interfaces[interface_type])
        if nso_to_oc_interface_types[interface_type] == "softwareLoopback":
            configure_software_loopback(config_before, config_leftover, interfaces[interface_type])
        if nso_to_oc_interface_types[interface_type] == "l3ipvlan":
            configure_software_l3ipvlan(config_before, config_leftover, interfaces[interface_type])
        if nso_to_oc_interface_types[interface_type] == "tunnel":
            configure_software_tunnel(config_before, config_leftover, interfaces[interface_type], if_ip)
        if nso_to_oc_interface_types[interface_type] == "vasi":
            configure_software_vasi(config_before, config_leftover, interfaces[interface_type])


def main(before: dict, leftover: dict, if_ip: dict, translation_notes: list = []) -> dict:
    """
    Translates NSO Device configurations to MDD OpenConfig configurations.

    Requires environment variables:
    NSO_URL: str
    NSO_USERNAME: str
    NSO_PASSWORD: str
    NSO_DEVICE: str
    TEST - If True, sends generated OC configuration to NSO Server: str

    :param before: Original NSO Device configuration: dict
    :param leftover: NSO Device configuration minus configs replaced with MDD OC: dict
    :param if_ip: Map of interface names to IP addresses: dict
    :return: MDD Openconfig Interfaces configuration: dict
    """

    interfaces = create_interface_dict(before)
    xe_interfaces(before, leftover, interfaces, if_ip)
    translation_notes += interfaces_notes

    return openconfig_interfaces


if __name__ == "__main__":
    sys.path.append("../../")
    sys.path.append("../../../")

    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xe import common_xe
        from package_nso_to_oc import common
    else:
        import common_xe
        import common

    (config_before_dict, config_leftover_dict, interface_ip_dict) = common_xe.init_xe_configs()
    main(config_before_dict, config_leftover_dict, interface_ip_dict)
    config_name = "_interfaces"
    config_remaining_name = "_remaining_interfaces"
    oc_name = "_openconfig_interfaces"
    common.print_and_test_configs(
        "xe1", config_before_dict, config_leftover_dict, openconfig_interfaces,
        config_name, config_remaining_name, oc_name, interfaces_notes)
else:
    # This is needed for now due to top level __init__.py. We need to determine if contents in __init__.py is still necessary.
    if (find_spec("package_nso_to_oc") is not None):
        from package_nso_to_oc.xe import common_xe
        from package_nso_to_oc import common
    else:
        from xe import common_xe
        import common
